# Copyright 2016 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from neutron_lib.callbacks import registry
from neutron_lib import context as neutron_context
from neutron_lib import exceptions
from oslo_config import cfg
from oslo_log import log as logging

from vmware_nsx.common import utils as nsx_utils
from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import formatters
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsx.shell.admin.plugins.nsxv3.resources import utils
import vmware_nsx.shell.resources as shell

LOG = logging.getLogger(__name__)
neutron_client = utils.NeutronDbClient()


@admin_utils.output_header
def list_dhcp_bindings(resource, event, trigger, **kwargs):
    """List DHCP bindings in Neutron."""

    comp_ports = [port for port in neutron_client.get_ports()
                  if nsx_utils.is_port_dhcp_configurable(port)]
    LOG.info(formatters.output_formatter(constants.DHCP_BINDING, comp_ports,
                                         ['id', 'mac_address', 'fixed_ips']))


@admin_utils.output_header
def nsx_recreate_dhcp_server(resource, event, trigger, **kwargs):
    """Recreate DHCP server & binding for a neutron network"""
    if not cfg.CONF.nsx_v3.native_dhcp_metadata:
        LOG.error("Native DHCP is disabled.")
        return

    errmsg = ("Need to specify net-id property. Add --property net-id=<id>")
    if not kwargs.get('property'):
        LOG.error("%s", errmsg)
        return
    properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
    net_id = properties.get('net-id')
    if not net_id:
        LOG.error("%s", errmsg)
        return

    context = neutron_context.get_admin_context()
    with utils.NsxV3PluginWrapper() as plugin:
        # verify that this is an existing network with dhcp enabled
        try:
            network = plugin._get_network(context, net_id)
        except exceptions.NetworkNotFound:
            LOG.error("Network %s was not found", net_id)
            return
        if plugin._has_no_dhcp_enabled_subnet(context, network):
            LOG.error("Network %s has no DHCP enabled subnet", net_id)
            return
        dhcp_relay = plugin.get_network_az_by_net_id(
            context, net_id).dhcp_relay_service
        if dhcp_relay:
            LOG.error("Native DHCP should not be enabled with dhcp relay")
            return

        # find the dhcp subnet of this network
        subnet_id = None
        for subnet in network.subnets:
            if subnet.enable_dhcp:
                subnet_id = subnet.id
                break
        if not subnet_id:
            LOG.error("Network %s has no DHCP enabled subnet", net_id)
            return
        dhcp_subnet = plugin.get_subnet(context, subnet_id)
        # disable and re-enable the dhcp
        plugin._enable_native_dhcp(context, network, dhcp_subnet)
    LOG.info("Done.")


registry.subscribe(list_dhcp_bindings,
                   constants.DHCP_BINDING,
                   shell.Operations.LIST.value)
registry.subscribe(nsx_recreate_dhcp_server,
                   constants.DHCP_BINDING,
                   shell.Operations.NSX_RECREATE.value)



# Copyright 2020 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os
import random
import re

from oslo_config import cfg
from oslo_context import context as context_utils
from oslo_log import log as logging
from oslo_utils import fileutils

from neutron import version as n_version
from neutron_lib.api import validators
from neutron_lib import context as q_context

from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.plugins.nsx_v3 import cert_utils
from vmware_nsxlib import v3
from vmware_nsxlib.v3 import client_cert
from vmware_nsxlib.v3 import config
from vmware_nsxlib.v3 import policy

LOG = logging.getLogger(__name__)

OS_NEUTRON_ID_SCOPE = 'os-neutron-id'
PORT_SG_SCOPE = 'os-security-group'

NSX_NEUTRON_PLUGIN = 'NSX Neutron plugin'


def get_DbCertProvider(conf_path):
    class DbCertProvider(client_cert.ClientCertProvider):
        """Write cert data from DB to file and delete after use

           New provider object with random filename is created for each
           request.
           This is not most efficient, but the safest way to avoid race
           conditions, since backend connections can occur both before and
           after neutron fork, and several concurrent requests can occupy the
           same thread.
           Note that new cert filename for each request does not result in new
           connection for each request (at least for now..)
        """
        EXPIRATION_ALERT_DAYS = 30          # days prior to expiration

        def __init__(self):
            super(DbCertProvider, self).__init__(None)
            random.seed()
            self._filename = '/tmp/.' + str(random.randint(1, 10000000))
            self.conf_path = conf_path

        def _check_expiration(self, expires_in_days):
            if expires_in_days > self.EXPIRATION_ALERT_DAYS:
                return

            if expires_in_days < 0:
                LOG.error("Client certificate has expired %d days ago.",
                          expires_in_days * -1)
            else:
                LOG.warning("Client certificate expires in %d days. "
                            "Once expired, service will become unavailable.",
                            expires_in_days)

        def __enter__(self):
            try:
                context = q_context.get_admin_context()
                db_storage_driver = cert_utils.DbCertificateStorageDriver(
                    context, self.conf_path.nsx_client_cert_pk_password)
                with client_cert.ClientCertificateManager(
                    cert_utils.NSX_OPENSTACK_IDENTITY,
                    None,
                    db_storage_driver) as cert_manager:
                    if not cert_manager.exists():
                        msg = _("Unable to load from nsx-db")
                        raise nsx_exc.ClientCertificateException(err_msg=msg)

                    filename = self._filename
                    if not os.path.exists(os.path.dirname(filename)):
                        if len(os.path.dirname(filename)) > 0:
                            fileutils.ensure_tree(os.path.dirname(filename))
                    cert_manager.export_pem(filename)

                    expires_in_days = cert_manager.expires_in_days()
                    self._check_expiration(expires_in_days)
            except Exception as e:
                self._on_exit()
                raise e

            return self

        def _on_exit(self):
            if os.path.isfile(self._filename):
                os.remove(self._filename)

            self._filename = None

        def __exit__(self, type, value, traceback):
            self._on_exit()

        def filename(self):
            return self._filename

    return DbCertProvider


def get_client_cert_provider(conf_path=cfg.CONF.nsx_v3):
    if not conf_path.nsx_use_client_auth:
        return None

    if conf_path.nsx_client_cert_storage.lower() == 'none':
        # Admin is responsible for providing cert file, the plugin
        # should not touch it
        return client_cert.ClientCertProvider(
                conf_path.nsx_client_cert_file)

    if conf_path.nsx_client_cert_storage.lower() == 'nsx-db':
        # Cert data is stored in DB, and written to file system only
        # when new connection is opened, and deleted immediately after.
        return get_DbCertProvider(conf_path)

    return None


def get_nsxlib_wrapper(nsx_username=None, nsx_password=None, basic_auth=False,
                       plugin_conf=None, allow_overwrite_header=False,
                       retriable_exceptions=None):
    if not plugin_conf:
        plugin_conf = cfg.CONF.nsx_v3

    client_cert_provider = None
    if not basic_auth:
        # if basic auth requested, dont use cert file even if provided
        client_cert_provider = get_client_cert_provider(conf_path=plugin_conf)

    exception_config = config.ExceptionConfig()
    if retriable_exceptions:
        exception_config.retriables = retriable_exceptions
    nsxlib_config = config.NsxLibConfig(
        username=nsx_username or plugin_conf.nsx_api_user,
        password=nsx_password or plugin_conf.nsx_api_password,
        client_cert_provider=client_cert_provider,
        retries=plugin_conf.http_retries,
        insecure=plugin_conf.insecure,
        ca_file=plugin_conf.ca_file,
        concurrent_connections=plugin_conf.concurrent_connections,
        http_timeout=plugin_conf.http_timeout,
        http_read_timeout=plugin_conf.http_read_timeout,
        conn_idle_timeout=plugin_conf.conn_idle_timeout,
        http_provider=None,
        max_attempts=plugin_conf.retries,
        nsx_api_managers=plugin_conf.nsx_api_managers,
        plugin_scope=OS_NEUTRON_ID_SCOPE,
        plugin_tag=NSX_NEUTRON_PLUGIN,
        plugin_ver=n_version.version_info.release_string(),
        dns_nameservers=cfg.CONF.nsx_v3.nameservers,
        dns_domain=cfg.CONF.nsx_v3.dns_domain,
        allow_overwrite_header=allow_overwrite_header,
        exception_config=exception_config)
    return v3.NsxLib(nsxlib_config)


def get_nsxpolicy_wrapper(nsx_username=None, nsx_password=None,
                          basic_auth=False, conf_path=None,
                          retriable_exceptions=None):
    if not conf_path:
        conf_path = cfg.CONF.nsx_p
    client_cert_provider = None
    if not basic_auth:
        # if basic auth requested, dont use cert file even if provided
        client_cert_provider = get_client_cert_provider(
            conf_path=conf_path)

    exception_config = config.ExceptionConfig()
    if retriable_exceptions:
        exception_config.retriables = retriable_exceptions

    nsxlib_config = config.NsxLibConfig(
        username=nsx_username or conf_path.nsx_api_user,
        password=nsx_password or conf_path.nsx_api_password,
        client_cert_provider=client_cert_provider,
        retries=conf_path.http_retries,
        insecure=conf_path.insecure,
        ca_file=conf_path.ca_file,
        concurrent_connections=conf_path.concurrent_connections,
        http_timeout=conf_path.http_timeout,
        http_read_timeout=conf_path.http_read_timeout,
        conn_idle_timeout=conf_path.conn_idle_timeout,
        http_provider=None,
        max_attempts=conf_path.retries,
        nsx_api_managers=conf_path.nsx_api_managers,
        plugin_scope=OS_NEUTRON_ID_SCOPE,
        plugin_tag=NSX_NEUTRON_PLUGIN,
        plugin_ver=n_version.version_info.release_string(),
        dns_nameservers=conf_path.nameservers,
        dns_domain=conf_path.dns_domain,
        exception_config=exception_config,
        allow_passthrough=(conf_path.allow_passthrough
                           if hasattr(conf_path, 'allow_passthrough')
                           else False),
        realization_max_attempts=(conf_path.realization_max_attempts
                                  if hasattr(conf_path,
                                             'realization_max_attempts')
                                  else 50),
        realization_wait_sec=(conf_path.realization_wait_sec
                              if hasattr(conf_path, 'realization_wait_sec')
                              else 1))
    return policy.NsxPolicyLib(nsxlib_config)


def inject_headers():
    ctx = context_utils.get_current()
    if ctx:
        ctx_dict = ctx.to_dict()
        # Remove unsupported characters from the user-id
        user_id = ctx_dict.get('user_identity')
        re.sub('[^A-Za-z0-9]+', '', user_id)
        return {'X-NSX-EUSER': user_id,
                'X-NSX-EREQID': ctx_dict.get('request_id')}
    return {}


def get_network_dns_domain(az, network):
    dns_domain = None
    if network.get('dns_domain'):
        net_dns = network['dns_domain']
        if isinstance(net_dns, str):
            dns_domain = net_dns
        elif hasattr(net_dns, "dns_domain"):
            dns_domain = net_dns.dns_domain
    if not dns_domain or not validators.is_attr_set(dns_domain):
        dns_domain = az.dns_domain
    return dns_domain


def build_dhcp_server_config(nsxlib, project_name, network, subnet, port, az):

    name = nsxlib.native_dhcp.build_server_name(
        network['name'], network['id'])

    net_tags = nsxlib.build_v3_tags_payload(
        network, resource_type='os-neutron-net-id',
        project_name=project_name)

    dns_domain = get_network_dns_domain(az, network)

    dns_nameservers = subnet['dns_nameservers']
    if not dns_nameservers or not validators.is_attr_set(dns_nameservers):
        dns_nameservers = az.nameservers

    # There must be exactly one fixed ip matching given subnet
    fixed_ip_addr = [fip['ip_address'] for fip in port['fixed_ips']
                     if fip['subnet_id'] == subnet['id']]
    return nsxlib.native_dhcp.build_server(
        name,
        ip_address=fixed_ip_addr[0],
        cidr=subnet['cidr'],
        gateway_ip=subnet['gateway_ip'],
        host_routes=subnet['host_routes'],
        dns_domain=dns_domain,
        dns_nameservers=dns_nameservers,
        dhcp_profile_id=az._native_dhcp_profile_uuid,
        tags=net_tags)



# Copyright 2018 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
from oslo_config import cfg

from neutron.db import l3_dvr_db  # noqa
from neutron import manager
from neutron_lib import context
from neutron_lib.plugins import constants as const
from neutron_lib.plugins import directory

from vmware_nsx.common import config
from vmware_nsx.plugins.common_v3 import utils as v3_utils
from vmware_nsx.plugins.nsx_p import plugin
from vmware_nsx.services.fwaas.nsx_p import fwaas_callbacks_v2
from vmware_nsx.shell.admin.plugins.common import formatters
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils

LOG = logging.getLogger(__name__)
_NSXPOLICY = None


def get_nsxp_client(nsx_username=None, nsx_password=None,
                    use_basic_auth=False):

    return get_connected_nsxpolicy(nsx_username,
                                   nsx_password,
                                   use_basic_auth).client


def get_connected_nsxpolicy(nsx_username=None, nsx_password=None,
                            use_basic_auth=False, conf_path=None,
                            retriable_exceptions=None,
                            verbose=False):
    global _NSXPOLICY

    if not verbose:
        # Suppress logs for nsxpolicy init
        logging.disable(logging.INFO)

    # for non-default arguments, initiate new lib
    if nsx_username or use_basic_auth:
        if not verbose:
            # Return logs to normal
            logging.disable(logging.NOTSET)
        return v3_utils.get_nsxpolicy_wrapper(
            nsx_username, nsx_password, use_basic_auth,
            conf_path=conf_path,
            retriable_exceptions=retriable_exceptions)
    if _NSXPOLICY is None:
        _NSXPOLICY = v3_utils.get_nsxpolicy_wrapper(
            conf_path=conf_path,
            retriable_exceptions=retriable_exceptions)
    if not verbose:
        # Return logs to normal
        logging.disable(logging.NOTSET)
    return _NSXPOLICY


def log_info(resource, data, attrs=['display_name', 'id']):
    LOG.info(formatters.output_formatter(resource, data, attrs))


def get_realization_info(resource, *realization_args):
    try:
        nsx_info = resource.get_realization_info(*realization_args,
                                                 silent=True)
        if not nsx_info:
            info_text = "MISSING"
        else:
            state = nsx_info.get('state')
            nsx_id = nsx_info.get('realization_specific_identifier')
            info_text = "%s (ID: %s)" % (state, nsx_id)
    except Exception as e:
        LOG.warning("Failed to get realization info for %s(%s): %s",
                    resource, str(realization_args), e)
        info_text = "UNKNOWN"
    return info_text


class NsxPolicyPluginWrapper(plugin.NsxPolicyPlugin):
    def __init__(self, verbose=False):
        if not verbose:
            # Suppress logs for plugin init
            logging.disable(logging.INFO)

        # initialize the availability zones
        config.register_nsxp_azs(cfg.CONF, cfg.CONF.nsx_p.availability_zones)
        super(NsxPolicyPluginWrapper, self).__init__()
        self.context = context.get_admin_context()
        admin_utils._init_plugin_mock_quota()

        if not verbose:
            # Return logs to normal
            logging.disable(logging.NOTSET)

    def __enter__(self):
        directory.add_plugin(const.CORE, self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        directory.add_plugin(const.CORE, None)

    def _init_fwaas_plugin(self, provider, callbacks_class, plugin_callbacks):
        fwaas_plugin_class = manager.NeutronManager.load_class_for_provider(
            'neutron.service_plugins', provider)
        fwaas_plugin = fwaas_plugin_class()
        self.fwaas_callbacks = callbacks_class(False)
        # override the fwplugin_rpc since there is no RPC support in adminutils
        if plugin_callbacks:
            self.fwaas_callbacks.fwplugin_rpc = plugin_callbacks(fwaas_plugin)
        self.init_is_complete = True

    def init_fwaas_for_admin_utils(self):
        # initialize the FWaaS plugin and callbacks
        self.fwaas_callbacks = None
        # This is an ugly patch to find out if fwaas is enabled
        service_plugins = cfg.CONF.service_plugins
        for srv_plugin in service_plugins:
            if 'firewall' in srv_plugin or 'fwaas' in srv_plugin:
                if 'v2' in srv_plugin:
                    # FWaaS V2
                    self._init_fwaas_plugin(
                        'firewall_v2',
                        fwaas_callbacks_v2.NsxpFwaasCallbacksV2,
                        None)
                return



# Copyright 2018 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import db_base_plugin_v2
from neutron.db import l3_db
from neutron_lib.callbacks import registry
from neutron_lib import context
from oslo_config import cfg
from oslo_log import log as logging

from vmware_nsx.db import nsx_models
from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsx.shell.admin.plugins.nsxp.resources import utils as p_utils
from vmware_nsx.shell import resources as shell

from vmware_nsxlib.v3 import nsx_constants
from vmware_nsxlib.v3.policy import constants as policy_constants
from vmware_nsxlib.v3.policy import transaction as policy_trans

LOG = logging.getLogger(__name__)


class RoutersPlugin(db_base_plugin_v2.NeutronDbPluginV2,
                    l3_db.L3_NAT_db_mixin):
    pass


@admin_utils.list_handler(constants.ROUTERS)
@admin_utils.output_header
def list_routers(resource, event, trigger, **kwargs):
    """List neutron routers

    With the NSX policy resources and realization state.
    """
    mappings = []
    nsxpolicy = p_utils.get_connected_nsxpolicy()
    ctx = context.get_admin_context()
    with p_utils.NsxPolicyPluginWrapper() as plugin:
        routers = plugin.get_routers(ctx, fields=['id', 'name', 'tenant_id'])
        for rtr in routers:
            status = p_utils.get_realization_info(
                nsxpolicy.tier1, rtr['id'])
            mappings.append({'ID': rtr['id'],
                             'Name': rtr.get('name'),
                             'Project': rtr.get('tenant_id'),
                             'NSX status': status})
    p_utils.log_info(constants.ROUTERS,
                     mappings,
                     attrs=['Project', 'Name', 'ID', 'NSX status'])
    return bool(mappings)


@admin_utils.output_header
def update_tier0(resource, event, trigger, **kwargs):
    """Replace old tier0 with a new one on the neutron DB and NSX backend"""
    errmsg = ("Need to specify old and new tier0 ID. Add --property "
              "old-tier0=<id> --property new-tier0=<id>")
    if not kwargs.get('property'):
        LOG.error("%s", errmsg)
        return
    properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
    old_tier0 = properties.get('old-tier0')
    new_tier0 = properties.get('new-tier0')
    if not old_tier0 or not new_tier0:
        LOG.error("%s", errmsg)
        return
    # Verify the id of the new tier0 (old one might not exist any more)
    nsxpolicy = p_utils.get_connected_nsxpolicy()
    try:
        nsxpolicy.tier0.get(new_tier0)
    except Exception:
        LOG.error("Tier0 logical router %s was not found", new_tier0)
        return

    # update all neutron DB entries
    old_tier0_networks = []
    ctx = context.get_admin_context()
    with ctx.session.begin(subtransactions=True):
        bindings = ctx.session.query(
            nsx_models.TzNetworkBinding).filter_by(phy_uuid=old_tier0).all()
        for bind in bindings:
            old_tier0_networks.append(bind.network_id)
            bind.phy_uuid = new_tier0

    if not old_tier0_networks:
        LOG.info("Did not find any provider networks using tier0 %s",
                 old_tier0)
        return

    LOG.info("Updated provider networks in DB: %s", old_tier0_networks)

    # Update tier1 routers GW to point to the new tier0 in the backend
    plugin = RoutersPlugin()
    neutron_routers = plugin.get_routers(ctx)
    for router in neutron_routers:
        router_gw_net = (router.get('external_gateway_info') and
                         router['external_gateway_info'].get('network_id'))
        if router_gw_net and router_gw_net in old_tier0_networks:
            try:
                nsxpolicy.tier1.update(router['id'], tier0=new_tier0)
            except Exception as e:
                LOG.error("Failed to update router %s linked port: %s",
                          router['id'], e)
            else:
                LOG.info("Updated router %s uplink port", router['id'])

    LOG.info("Done.")


@admin_utils.output_header
def update_nat_firewall_match(resource, event, trigger, **kwargs):
    """Update the firewall_match value in neutron nat rules with a new value"""
    errmsg = ("Need to specify internal/external firewall_match value. "
              "Add --property firewall-match=<match>")
    if not kwargs.get('property'):
        LOG.error("%s", errmsg)
        return
    properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
    firewall_match_str = properties.get('firewall-match')
    if (not firewall_match_str or
        firewall_match_str.lower() not in ('internal', 'external')):
        LOG.error("%s", errmsg)
        return

    if firewall_match_str.lower() == 'internal':
        new_firewall_match = policy_constants.NAT_FIREWALL_MATCH_INTERNAL
        old_firewall_match = policy_constants.NAT_FIREWALL_MATCH_EXTERNAL
        conf_match_internal = True
    else:
        new_firewall_match = policy_constants.NAT_FIREWALL_MATCH_EXTERNAL
        old_firewall_match = policy_constants.NAT_FIREWALL_MATCH_INTERNAL
        conf_match_internal = False

    cfg.CONF.set_override('firewall_match_internal_addr',
                          conf_match_internal, 'nsx_p')

    nsxpolicy = p_utils.get_connected_nsxpolicy()
    with p_utils.NsxPolicyPluginWrapper() as plugin:
        # Make sure FWaaS was initialized
        plugin.init_fwaas_for_admin_utils()

        ctx = context.get_admin_context()
        neutron_routers = plugin.get_routers(ctx)
        for router in neutron_routers:
            rules = nsxpolicy.tier1_nat_rule.list(router['id'])
            for rule in rules:
                if not nsxpolicy.feature_supported(
                    nsx_constants.FEATURE_PARTIAL_UPDATES):
                    if rule.get('firewall_match') == old_firewall_match:
                        nsxpolicy.tier1_nat_rule.update(
                            router['id'], rule['id'],
                            firewall_match=new_firewall_match)
                else:
                    with policy_trans.NsxPolicyTransaction():
                        if rule.get('firewall_match') == old_firewall_match:
                            nsxpolicy.tier1_nat_rule.update(
                                router['id'], rule['id'],
                                firewall_match=new_firewall_match)

            if plugin.fwaas_callbacks:
                # get all router interface networks
                interface_ports = plugin._get_router_interfaces(
                    ctx, router['id'])
                for port in interface_ports:
                    plugin.fwaas_callbacks.update_segment_group(
                        ctx, router['id'], port['network_id'])

    LOG.info("Done.")


registry.subscribe(update_tier0,
                   constants.ROUTERS,
                   shell.Operations.UPDATE_TIER0.value)

registry.subscribe(update_nat_firewall_match,
                   constants.ROUTERS,
                   shell.Operations.UPDATE_FIREWALL_MATCH.value)



# Copyright 2019 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

import netaddr
from oslo_log import log as logging

from neutron_lib.api.definitions import allowedaddresspairs as addr_apidef
from neutron_lib.api.definitions import provider_net as pnet
from neutron_lib.api import validators
from neutron_lib.callbacks import registry
from neutron_lib import constants as nl_constants
from neutron_lib import context as n_context

from vmware_nsx.common import nsxv_constants
from vmware_nsx.common import utils as c_utils
from vmware_nsx.services.lbaas.nsx_v3.implementation import lb_utils
from vmware_nsx.services.lbaas.octavia import constants as oct_const
from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsx.shell.admin.plugins.nsxv.resources import utils
from vmware_nsx.shell import resources as shell
from vmware_nsxlib.v3 import nsx_constants as nsxlib_consts

LOG = logging.getLogger(__name__)


@admin_utils.output_header
def validate_config_for_migration(resource, event, trigger, **kwargs):
    """Validate the nsxv configuration before migration to nsx-t"""

    transit_networks = ["100.64.0.0/16"]
    if kwargs.get('property'):
        # input validation
        properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
        transit_network = properties.get('transit-network')
        if transit_network:
            transit_networks = [transit_network]

    # Max number of allowed address pairs (allowing 3 for fixed ips)
    num_allowed_addr_pairs = nsxlib_consts.NUM_ALLOWED_IP_ADDRESSES - 3

    admin_context = n_context.get_admin_context()
    n_errors = 0

    with utils.NsxVPluginWrapper() as plugin:
        # Ports validations:
        ports = plugin.get_ports(admin_context)
        for port in ports:
            net_id = port['network_id']
            # Too many address pairs in a port
            address_pairs = port.get(addr_apidef.ADDRESS_PAIRS)
            if len(address_pairs) > num_allowed_addr_pairs:
                n_errors = n_errors + 1
                LOG.error("%s allowed address pairs for port %s. Only %s are "
                          "allowed.",
                          len(address_pairs), port['id'],
                          num_allowed_addr_pairs)

            # Compute port on external network
            if (port.get('device_owner', '').startswith(
                    nl_constants.DEVICE_OWNER_COMPUTE_PREFIX) and
                plugin._network_is_external(admin_context, net_id)):
                n_errors = n_errors + 1
                LOG.error("Compute port %s on external network %s is not "
                          "allowed.", port['id'], net_id)

        # Networks & subnets validations:
        networks = plugin.get_networks(admin_context)
        for net in networks:
            # skip internal networks
            if net['project_id'] == nsxv_constants.INTERNAL_TENANT_ID:
                continue

            # VXLAN or portgroup provider networks
            net_type = net.get(pnet.NETWORK_TYPE)
            if (net_type == c_utils.NsxVNetworkTypes.VXLAN or
                net_type == c_utils.NsxVNetworkTypes.PORTGROUP):
                n_errors = n_errors + 1
                LOG.error("Network %s of type %s is not supported.",
                          net['id'], net_type)

            subnets = plugin._get_subnets_by_network(admin_context, net['id'])
            n_dhcp_subnets = 0

            # Multiple DHCP subnets per network
            for subnet in subnets:
                if subnet['enable_dhcp']:
                    n_dhcp_subnets = n_dhcp_subnets + 1
            if n_dhcp_subnets > 1:
                n_errors = n_errors + 1
                LOG.error("Network %s has %s dhcp subnets. Only 1 is allowed.",
                          net['id'], n_dhcp_subnets)

            # Subnets overlapping with the transit network
            for subnet in subnets:
                # get the subnet IPs
                if ('allocation_pools' in subnet and
                    validators.is_attr_set(subnet['allocation_pools'])):
                    # use the pools instead of the cidr
                    subnet_networks = [
                        netaddr.IPRange(pool.get('start'), pool.get('end'))
                        for pool in subnet.get('allocation_pools')]
                else:
                    cidr = subnet.get('cidr')
                    if not validators.is_attr_set(cidr):
                        return
                    subnet_networks = [netaddr.IPNetwork(subnet['cidr'])]

                for subnet_net in subnet_networks:
                    if (netaddr.IPSet(subnet_net) &
                        netaddr.IPSet(transit_networks)):
                        n_errors = n_errors + 1
                        LOG.error("Subnet %s overlaps with the transit "
                                  "network ips: %s.",
                                  subnet['id'], transit_networks)

            # Network attached to multiple routers
            intf_ports = plugin._get_network_interface_ports(
                admin_context, net['id'])
            if len(intf_ports) > 1:
                n_errors = n_errors + 1
                LOG.error("Network %s has interfaces on multiple routers. "
                          "Only 1 is allowed.", net['id'])

        # Routers validations:
        routers = plugin.get_routers(admin_context)
        for router in routers:
            # Interface subnets overlap with the GW subnet
            gw_subnets = plugin._find_router_gw_subnets(admin_context, router)
            gw_cidrs = [subnet['cidr'] for subnet in gw_subnets]
            gw_ip_set = netaddr.IPSet(gw_cidrs)

            if_cidrs = plugin._find_router_subnets_cidrs(
                admin_context, router['id'])
            if_ip_set = netaddr.IPSet(if_cidrs)

            if gw_ip_set & if_ip_set:
                n_errors = n_errors + 1
                LOG.error("Interface network of router %s cannot overlap with "
                          "router GW network", router['id'])

        # TODO(asarfaty): missing validations:
        # - Vlan provider network with the same VLAN tag as the uplink
        #   profile tag used in the relevant transport node
        #   (cannot check this without access to the T manager)

        # Octavia loadbalancers validation:
        filters = {'device_owner': [nl_constants.DEVICE_OWNER_LOADBALANCERV2,
                                    oct_const.DEVICE_OWNER_OCTAVIA]}
        lb_ports = plugin.get_ports(admin_context, filters=filters)
        for port in lb_ports:
            fixed_ips = port.get('fixed_ips', [])
            if fixed_ips:
                subnet_id = fixed_ips[0]['subnet_id']
                network = lb_utils.get_network_from_subnet(
                    admin_context, plugin, subnet_id)
                router_id = lb_utils.get_router_from_network(
                    admin_context, plugin, subnet_id)
                # Loadbalancer vip subnet must be connected to a router or
                # belong to an external network
                if (not router_id and network and
                    not network.get('router:external')):
                    n_errors = n_errors + 1
                    LOG.error("Loadbalancer %s subnet %s is not external "
                              "nor connected to a router.",
                              port.get('device_id'), subnet_id)

            # TODO(asarfaty): Multiple listeners on the same pool is not
            # supported, but currently the admin utility has no access to this
            # information from octavia

            # TODO(asarfaty): Member on external subnet must have fip as ip,
            # but currently the admin utility has no access to this information
            # from octavia

    if n_errors > 0:
        plural = n_errors > 1
        LOG.error("The NSX-V plugin configuration is not ready to be "
                  "migrated to NSX-T. %s error%s found.", n_errors,
                  's were' if plural else ' was')
        sys.exit(n_errors)

    LOG.info("The NSX-V plugin configuration is ready to be migrated to "
             "NSX-T.")


registry.subscribe(validate_config_for_migration,
                   constants.NSX_MIGRATE_V_T,
                   shell.Operations.VALIDATE.value)



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This file contains FWaaS mocks, to allow the vmware nsx plugins to work when
# FWaaS code does not exist, and FWaaS is not configured in neutron

FIREWALL_V2 = 'FIREWALL_V2'


class L3WithFWaaS(object):
    def __init__(self, **kwargs):
        self.fwaas_enabled = False


class FwaasDriverBase(object):
    pass


class FirewallPluginV2(object):
    pass


class FirewallCallbacks(object):
    pass



# Copyright 2020 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import sys
import time

import logging
import paramiko
import tenacity

from neutron.extensions import securitygroup as ext_sg
from neutron_fwaas.db.firewall.v2 import firewall_db_v2
from neutron_lib.callbacks import registry
from neutron_lib import context
from oslo_config import cfg

from vmware_nsx.common import utils as nsx_utils
from vmware_nsx.db import db
from vmware_nsx.db import nsx_models
from vmware_nsx.plugins.nsx_p import plugin as p_plugin
from vmware_nsx.plugins.nsx_v3 import cert_utils
from vmware_nsx.plugins.nsx_v3 import plugin as v3_plugin
from vmware_nsx.plugins.nsx_v3 import utils as v3_plugin_utils
from vmware_nsx.services.fwaas.nsx_p import fwaas_callbacks_v2
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_p.implementation import lb_utils
from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsx.shell.admin.plugins.nsxp.resources import utils as p_utils
from vmware_nsx.shell.admin.plugins.nsxv3.resources import utils
from vmware_nsx.shell import resources as shell

from vmware_nsxlib.v3 import core_resources as nsx_resources
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3 import load_balancer as nsxlib_lb
from vmware_nsxlib.v3 import nsx_constants
from vmware_nsxlib.v3.policy import constants as policy_constants
from vmware_nsxlib.v3.policy import core_resources as policy_resources
from vmware_nsxlib.v3.policy import utils as policy_utils

LOG = logging.getLogger(__name__)

POLICY_API_STATUS_FAILED = 'FAILED'
POLICY_API_STATUS_SUCCESS = 'SUCCESS'
POLICY_API_STATUS_IN_PROGRESS = 'PAUSING'
POLICY_API_STATUS_PAUSED = 'PAUSED'
POLICY_API_STATUS_READY = 'NOT_STARTED'

STATUS_ALLOW_MIGRATION_REQ = set([
    POLICY_API_STATUS_SUCCESS,
    POLICY_API_STATUS_READY
])

MIGRATE_LIMIT_NO_LIMIT = 0
MIGRATE_LIMIT_TIER0 = 1
MIGRATE_LIMIT_TIER0_PORTS = 1000
MIGRATE_LIMIT_TIER1 = 1000
MIGRATE_LIMIT_TIER1_PORTS = 1000
MIGRATE_LIMIT_NAT = 1500
MIGRATE_LIMIT_DHCP_SERVER = 1500
MIGRATE_LIMIT_MD_PROXY = 1500
MIGRATE_LIMIT_SWITCH_PROFILE = 1500
MIGRATE_LIMIT_LOGICAL_SWITCH = 500
MIGRATE_LIMIT_LOGICAL_PORT = 1500
MIGRATE_LIMIT_NS_GROUP = 2000
MIGRATE_LIMIT_DFW_SECTION = 3000
MIGRATE_LIMIT_LB_SERVICE = 2000
MIGRATE_LIMIT_LB_VIRTUAL_SERVER = 2000
MIGRATE_LIMIT_LB_MONITOR = 1500
MIGRATE_LIMIT_LB_POOL = 1500
MIGRATE_LIMIT_LB_APP_PROFILE = 2000
MIGRATE_LIMIT_LB_PER_PROFILE = 2000
MIGRATE_LIMIT_CERT = 1500

COMPONENT_STATUS_ALREADY_MIGRATED = 1
COMPONENT_STATUS_OK = 2

ROLLBACK_DATA = []
EDGE_FW_SEQ = 1
DFW_SEQ = 1
NSX_ROUTER_SECTIONS = []
SERVICE_UP_RETRIES = 30


def start_migration_process(nsxlib):
    """Notify the manager that the migration process is starting"""
    return nsxlib.client.url_post(
        "migration/mp-to-policy/workflow?action=INITIATE", None)


def end_migration_process(nsxlib):
    """Notify the manager that the migration process has ended"""
    return nsxlib.client.url_post(
        "migration/mp-to-policy/workflow?action=DONE", None)


def send_migration_request(nsxlib, body):
    return nsxlib.client.url_post("migration/mp-to-policy", body)


def send_rollback_request(nsxlib, body):
    #TODO(asarfaty): Rollback can take very long, especially for firewall
    # sections. In this case backup-restore might be better
    return nsxlib.client.url_post("migration/mp-to-policy/rollback", body)


def send_migration_plan_action(nsxlib, action):
    return nsxlib.client.url_post("migration/plan?action=%s" % action, None)


def get_migration_status(nsxlib, silent=False):
    return nsxlib.client.get("migration/status-summary",
                             silent=silent)


def change_migration_service_status(start=True, nsxlib=None):
    """Enable/Disable the migration service on the NSX manager
    using SSH command
    """
    # TODO(asarfaty): Is there an api for that? or use sshpass
    action = 'start' if start else 'stop'
    command = "%s service migration-coordinator" % action
    LOG.info("Going to %s the migration service on the NSX manager by "
             "SSHing the manager and running '%s'", action, command)
    host = cfg.CONF.nsx_v3.nsx_api_managers[0]
    user = cfg.CONF.nsx_v3.nsx_api_user[0]
    passwd = cfg.CONF.nsx_v3.nsx_api_password[0]

    ssh = paramiko.SSHClient()
    ssh.load_system_host_keys()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(host, username=user, password=passwd)
    ssh.exec_command(command)

    if start and nsxlib:
        LOG.info("Waiting for the service to be up...")
        start_time = time.time()

        @tenacity.retry(reraise=True,
                        retry=tenacity.retry_if_exception_type(Exception),
                        wait=tenacity.wait_exponential(multiplier=0.5, max=2),
                        stop=tenacity.stop_after_attempt(SERVICE_UP_RETRIES))
        def get_migration_status_with_retry(nsxlib):
            get_migration_status(nsxlib, silent=True)

        try:
            get_migration_status_with_retry(nsxlib)
        except Exception:
            raise Exception("The migration service did not get up after %s "
                            "retries" % SERVICE_UP_RETRIES)

        elapsed_time = time.time() - start_time
        LOG.info("The service is up (waited %s seconds)", elapsed_time)


def ensure_migration_state_ready(nsxlib, with_abort=False):
    try:
        status = get_migration_status(nsxlib, silent=True)
    except Exception as e:
        if with_abort:
            change_migration_service_status(start=True, nsxlib=nsxlib)
            return ensure_migration_state_ready(nsxlib)
        LOG.debug("Failed to get migration status: %s", e)
        return False

    if status["overall_migration_status"] not in STATUS_ALLOW_MIGRATION_REQ:
        LOG.error("Migration status not ready: %s", status)
        if with_abort:
            send_migration_plan_action(nsxlib, 'abort')
            return ensure_migration_state_ready(
                nsxlib, with_abort=with_abort)
        return False

    return True


def verify_component_status(nsxlib, component_number):
    status = get_migration_status(nsxlib)
    if (status['component_status'][component_number]['status'] ==
            POLICY_API_STATUS_FAILED):
        # If it's a duplicate migration request, pass the verification
        if ('is duplicate or already migrated' in
                status['component_status'][component_number]['details'] and
            component_number == 0):
            # Success that indicates resource migration is already done
            return COMPONENT_STATUS_ALREADY_MIGRATED
        # bad state. abort, mark as fail, and go to next request
        raise Exception("The migration server returned with FAILURE status. "
                        "Details: %s", status)
    # Success
    return COMPONENT_STATUS_OK


def wait_on_overall_migration_status_to_pause(nsxlib):
    while True:
        status = get_migration_status(nsxlib)
        migration_status = status.get('overall_migration_status')
        if (migration_status == POLICY_API_STATUS_PAUSED or
            migration_status == POLICY_API_STATUS_SUCCESS):
            break
        time.sleep(1)


def printable_resource_name(resource):
    name = resource.get('display_name')
    if name:
        try:
            name = str(name)
        except UnicodeEncodeError:
            name = name.encode('ascii', 'ignore')
    res_id = resource.get('id')
    if name == res_id:
        return name
    return "%s (%s)" % (name, resource.get('id'))


def get_resource_migration_data(nsxlib_resource, neutron_id_tags,
                                resource_type, resource_condition=None,
                                printable_name=None, policy_resource_get=None,
                                policy_id_callback=None,
                                metadata_callback=None,
                                skip_policy_path_check=False,
                                nsxlib_list_args=None):
    if not printable_name:
        printable_name = resource_type
    LOG.debug("Getting data for MP %s", printable_name)

    if nsxlib_list_args:
        resources = nsxlib_resource.list(**nsxlib_list_args)
    else:
        resources = nsxlib_resource.list()
    if not isinstance(resources, list):
        # The nsxlib resources list return inconsistent type of result
        resources = resources.get('results', [])
    policy_ids = []
    entries = []
    for resource in resources:
        name_and_id = printable_resource_name(resource)
        policy_id = resource['id']
        # Go over tags and find the neutron id
        neutron_id = None
        found_policy_path = False
        for tag in resource.get('tags', []):
            if tag['scope'] == 'policyPath':
                # This is already a policy resource
                found_policy_path = True
            if neutron_id_tags and tag['scope'] in neutron_id_tags:
                neutron_id = tag['tag']
        if not skip_policy_path_check and found_policy_path:
            LOG.debug("Skipping %s %s as it is already a policy "
                      "resource", printable_name, name_and_id)
            continue
        if neutron_id_tags:
            if not neutron_id:
                # Not a neutron resource
                LOG.debug("Skipping %s %s as it is not a neutron resource",
                          printable_name, name_and_id)
                continue
            policy_id = neutron_id
        if resource_condition:
            if not resource_condition(resource):
                LOG.debug("Skipping %s %s as it does not match the neutron "
                          "condition", printable_name, name_and_id)
                continue
        if policy_id_callback:
            # Callback to change the policy id
            policy_id = policy_id_callback(resource, policy_id)
        if policy_id and policy_resource_get:
            # filter out resources that already exit on policy!
            try:
                policy_resource_get(policy_id, silent=True)
            except nsxlib_exc.ResourceNotFound:
                pass
            else:
                LOG.debug("Skipping %s %s as it already exists on the "
                          "policy backend", printable_name, name_and_id)
                continue

        # Make sure not to migrate multiple resources to the same policy-id
        if policy_id:
            if policy_id in policy_ids:
                msg = (_("Cannot migrate %s %s to policy-id %s: Another %s "
                         "has the same designated policy-id. One of those is "
                         "probably a neutron orphaned. Please delete it and "
                         "try migration again.") % (printable_name,
                         name_and_id, policy_id, printable_name))
                raise Exception(msg)
            policy_ids.append(policy_id)

        LOG.debug("Adding data for %s %s, policy-id %s",
                  printable_name, name_and_id, policy_id)
        entry = {'manager_id': resource['id']}
        if policy_id:
            entry['policy_id'] = policy_id
        if metadata_callback:
            metadata_callback(entry, policy_id, resource)
        entries.append(entry)
    return entries


def migrate_objects(nsxlib, data, use_admin=False):
    if not ensure_migration_state_ready(nsxlib):
        raise Exception("The migration server is not ready")

    migration_body = {"migration_data": [data]}

    # Update the principal identity for the policy resources
    # use 'admin' for predefined objects, and the opestack configured
    # user/identity for openstack resources
    if use_admin:
        user = 'admin'
    elif cfg.CONF.nsx_v3.nsx_use_client_auth:
        user = cert_utils.NSX_OPENSTACK_IDENTITY
    else:
        user = cfg.CONF.nsx_v3.nsx_api_user[0]
    migration_body['setup_details'] = {
        'principal_identity': user}

    LOG.info("Migrating %d %s objects with principal_identity %s",
             len(data['resource_ids']), data['type'], user)
    LOG.debug("Migration body : %s", migration_body)

    send_migration_request(nsxlib, migration_body)
    # send the start action
    send_migration_plan_action(nsxlib, 'start')

    # wait until the overall_migration_status is SUCCESS
    wait_on_overall_migration_status_to_pause(nsxlib)

    # verify first component status
    success_code = verify_component_status(nsxlib, 0)
    if success_code == COMPONENT_STATUS_ALREADY_MIGRATED:
        return True

    # send the continue action
    send_migration_plan_action(nsxlib, 'continue')

    # wait until the overall_migration_status is SUCCESS
    wait_on_overall_migration_status_to_pause(nsxlib)

    # verify second component status (Will raise in case of error)
    try:
        verify_component_status(nsxlib, 1)
    except Exception as e:
        raise e
    else:
        global ROLLBACK_DATA
        # rollback should be done in the reverse order
        ROLLBACK_DATA = [data] + ROLLBACK_DATA

    return True


def migrate_resource(nsxlib, resource_type, entries,
                     limit=MIGRATE_LIMIT_NO_LIMIT,
                     count_internals=False, use_admin=False):
    # Call migrate_resource with the part of resources we need by the limit
    if not entries:
        LOG.info("No %s to migrate", resource_type)
        return

    LOG.info("Going to migrate %d %s objects in groups of max %s",
             len(entries), resource_type, limit)
    start_time = time.time()

    if limit == MIGRATE_LIMIT_NO_LIMIT:
        migrate_objects(nsxlib, {'type': resource_type,
                                 'resource_ids': entries},
                        use_admin=use_admin)
    else:
        if count_internals:
            # Limit the total number of resources, including internal ones
            counter = 0
            entries_to_migrate = []
            for index in range(0, len(entries)):
                addition_size = 1 + len(entries[index].get('linked_ids', []))
                if addition_size > limit:
                    # Unsupported size of resource
                    raise Exception("%s size is over the allowed limit of "
                                    "%s" % (resource_type, limit))
                if counter + addition_size > limit:
                    # Migrate what was accumulated so far
                    migrate_objects(nsxlib,
                                    {'type': resource_type,
                                     'resource_ids': entries_to_migrate},
                                    use_admin=use_admin)
                    # Start a new accumulation
                    counter = addition_size
                    entries_to_migrate = [entries[index]]
                else:
                    # Keep accumulating
                    counter = counter + addition_size
                    entries_to_migrate.append(entries[index])
            if entries_to_migrate:
                # Migrate the left overs
                migrate_objects(nsxlib,
                                {'type': resource_type,
                                 'resource_ids': entries_to_migrate},
                                use_admin=use_admin)
        else:
            for index in range(0, len(entries), limit):
                migrate_objects(nsxlib,
                                {'type': resource_type,
                                 'resource_ids': entries[index:index + limit]},
                                use_admin=use_admin)

    elapsed_time = time.time() - start_time
    LOG.info("Migrating %d %s objects took %s seconds",
             len(entries), resource_type, elapsed_time)


def get_configured_values(plugin, az_attribute):
    values = []
    for az in plugin.get_azs_list():
        az_values = getattr(az, az_attribute)
        if isinstance(az_values, list):
            values.extend(az_values)
        else:
            values.append(az_values)
    return values


def get_neurton_tier0s(plugin):
    return get_configured_values(plugin, '_default_tier0_router')


def migrate_tier0s(nsxlib, nsxpolicy, plugin):
    # First prepare a list of neutron related tier0s from the config
    neutron_t0s = get_neurton_tier0s(plugin)
    # Add tier0s used specifically in external networks
    ctx = context.get_admin_context()
    with ctx.session.begin(subtransactions=True):
        bindings = ctx.session.query(
            nsx_models.TzNetworkBinding).filter_by(
            binding_type='l3_ext').all()
        for bind in bindings:
            if bind.phy_uuid not in neutron_t0s:
                neutron_t0s.append(bind.phy_uuid)

    def cond(resource):
        return (resource.get('router_type', '') == 'TIER0' and
                resource.get('id') in neutron_t0s)

    entries = get_resource_migration_data(
        nsxlib.logical_router, None,
        'TIER0', resource_condition=cond,
        policy_resource_get=nsxpolicy.tier0.get,
        nsxlib_list_args={'router_type': nsx_constants.ROUTER_TYPE_TIER0})
    migrate_resource(nsxlib, 'TIER0', entries, MIGRATE_LIMIT_TIER0,
                     use_admin=True)
    migrated_tier0s = [entry['manager_id'] for entry in entries]

    # Create a list of public switches connected to the tier0s to migrate later
    public_switches = []
    for tier0 in neutron_t0s:
        uplink_port = nsxlib.logical_router_port.get_tier0_uplink_port(tier0)
        if uplink_port:
            # Get the external LS id from the uplink port
            port_id = uplink_port['linked_logical_switch_port_id']['target_id']
            port = nsxlib.logical_port.get(port_id)
            public_switches.append(port['logical_switch_id'])

    return public_switches, migrated_tier0s


def is_neutron_resource(resource):
    # Return True if the resource has the neutron marking tag
    for tag in resource.get('tags', []):
        if tag.get('scope') == 'os-api-version':
            return True
    return False


def migrate_switch_profiles(nsxlib, nsxpolicy, plugin):
    """Return all types of neutron switching profiles"""

    # Build a condition for each type of switching profiles.
    # Note(asarfaty): system owned profiles should also be migrated as they are
    # missing from policy

    # Include switch profiles that are in the nsx.ini
    conf_profiles = get_configured_values(plugin, 'switching_profiles')

    # Add other switch profiles that might be used by neutron ports in the past
    port_profiles = set()
    ports = nsxlib.logical_port.list()['results']
    for port in ports:
        # Check that it is a neutron port
        if not is_neutron_resource(port):
            continue
        for prof in port.get('switching_profile_ids', []):
            port_profiles.add(prof['value'])

    # Black list neuron & system profiles that should not be migrated
    names_black_list = [v3_plugin_utils.NSX_V3_DHCP_PROFILE_NAME,
                        'ServiceInsertion_MacManagement_Profile']

    def get_cond(resource_type):
        def cond(resource):
            return (resource.get('resource_type') == resource_type and
                    resource.get('display_name') not in names_black_list and
                    (resource.get('id') in conf_profiles or
                     resource.get('id') in port_profiles or
                     resource.get('_system_owned', True) or
                     is_neutron_resource(resource)))
        return cond

    def get_policy_id_callback(res, policy_id):
        # In case of plugin init profiles: give it the id the policy plugin
        # will use
        mapping = {v3_plugin.NSX_V3_MAC_LEARNING_PROFILE_NAME:
                   p_plugin.MAC_DISCOVERY_PROFILE_ID,
                   v3_plugin_utils.NSX_V3_PSEC_PROFILE_NAME:
                   p_plugin.SPOOFGUARD_PROFILE_ID}

        if mapping.get(res.get('display_name')):
            return mapping[res['display_name']]

        # QoS profiles should get the neutron policy id
        for tag in res.get('tags', []):
            if tag['scope'] == 'os-neutron-qos-id':
                policy_id = tag['tag']

        return policy_id

    entries = get_resource_migration_data(
        nsxlib.switching_profile, None,
        'SPOOFGUARD_PROFILES',
        resource_condition=get_cond(
            nsx_resources.SwitchingProfileTypes.SPOOF_GUARD),
        policy_resource_get=nsxpolicy.spoofguard_profile.get,
        policy_id_callback=get_policy_id_callback)
    migrate_resource(nsxlib, 'SPOOFGUARD_PROFILES', entries,
                     MIGRATE_LIMIT_SWITCH_PROFILE)

    entries = get_resource_migration_data(
        nsxlib.switching_profile, None,
        'MACDISCOVERY_PROFILES',
        resource_condition=get_cond(
            nsx_resources.SwitchingProfileTypes.MAC_LEARNING),
        policy_resource_get=nsxpolicy.mac_discovery_profile.get,
        policy_id_callback=get_policy_id_callback)
    migrate_resource(nsxlib, 'MACDISCOVERY_PROFILES', entries,
                     MIGRATE_LIMIT_SWITCH_PROFILE)

    entries = get_resource_migration_data(
        nsxlib.switching_profile, None,
        'SEGMENT_SECURITY_PROFILES',
        resource_condition=get_cond(
            nsx_resources.SwitchingProfileTypes.SWITCH_SECURITY),
        policy_resource_get=nsxpolicy.segment_security_profile.get,
        policy_id_callback=get_policy_id_callback)
    migrate_resource(nsxlib, 'SEGMENT_SECURITY_PROFILES', entries,
                     MIGRATE_LIMIT_SWITCH_PROFILE)

    entries = get_resource_migration_data(
        nsxlib.switching_profile, None,
        'QOS_PROFILES',
        resource_condition=get_cond(
            nsx_resources.SwitchingProfileTypes.QOS),
        policy_resource_get=nsxpolicy.qos_profile.get,
        policy_id_callback=get_policy_id_callback)
    migrate_resource(nsxlib, 'QOS_PROFILES', entries,
                     MIGRATE_LIMIT_SWITCH_PROFILE)

    entries = get_resource_migration_data(
        nsxlib.switching_profile, None,
        'IPDISCOVERY_PROFILES',
        resource_condition=get_cond(
            nsx_resources.SwitchingProfileTypes.IP_DISCOVERY),
        policy_resource_get=nsxpolicy.ip_discovery_profile.get,
        policy_id_callback=get_policy_id_callback)
    migrate_resource(nsxlib, 'IPDISCOVERY_PROFILES', entries,
                     MIGRATE_LIMIT_SWITCH_PROFILE)


def migrate_md_proxies(nsxlib, nsxpolicy, plugin):
    neutron_md = get_configured_values(plugin, '_native_md_proxy_uuid')

    # Add other mdproxies that might be used by neutron networks in the past
    ports = nsxlib.logical_port.list()['results']
    for port in ports:
        # Check that it is a neutron port
        if not is_neutron_resource(port):
            continue
        if (port.get('attachment') and
            port['attachment'].get('attachment_type') == 'METADATA_PROXY'):
            mdproxy_id = port['attachment'].get('id')
            if mdproxy_id not in neutron_md:
                neutron_md.append(port['attachment'].get('id'))

    # make sure to migrate all certificates used by those MD proxies
    certificates = []
    for md_id in neutron_md:
        md_resource = nsxlib.native_md_proxy.get(md_id)
        certificates.extend(md_resource.get('metadata_server_ca_ids', []))

    if certificates:
        def cert_cond(resource):
            return resource.get('id') in certificates

        entries = get_resource_migration_data(
            nsxlib.trust_management, None,
            'CERTIFICATE',
            resource_condition=cert_cond,
            policy_resource_get=nsxpolicy.certificate.get)
        migrate_resource(nsxlib, 'CERTIFICATE', entries,
                         MIGRATE_LIMIT_CERT)

    # Now migrate the MD proxies
    def cond(resource):
        return resource.get('id') in neutron_md

    entries = get_resource_migration_data(
        nsxlib.native_md_proxy, None,
        'METADATA_PROXY',
        resource_condition=cond,
        policy_resource_get=nsxpolicy.md_proxy.get)
    migrate_resource(nsxlib, 'METADATA_PROXY', entries,
                     MIGRATE_LIMIT_MD_PROXY, use_admin=True)


def migrate_networks(nsxlib, nsxpolicy, plugin, public_switches):

    # Get a list of nsx-net provider networks to migrate
    # Those networks have no tags, and should keep the same id in policy
    nsx_networks = []
    ctx = context.get_admin_context()
    with ctx.session.begin(subtransactions=True):
        bindings = ctx.session.query(
            nsx_models.TzNetworkBinding).filter_by(
            binding_type=nsx_utils.NsxV3NetworkTypes.NSX_NETWORK).all()
        for bind in bindings:
            nsx_networks.append(bind.phy_uuid)

    def cond(resource):
        return (resource.get('id', '') in nsx_networks or
                resource.get('id', '') in public_switches or
                is_neutron_resource(resource))

    def get_policy_id(resource, policy_id):
        if resource['id'] in nsx_networks:
            # Keep original ID
            return resource['id']
        if resource['id'] in public_switches:
            # Keep original ID
            return resource['id']
        for tag in resource.get('tags', []):
            # Use the neutron ID
            if tag['scope'] == 'os-neutron-net-id':
                return tag['tag']

    def add_metadata(entry, policy_id, resource):
        # Add dhcp-v4 static bindings
        network_id = None
        for tag in resource.get('tags', []):
            # Use the neutron ID
            if tag['scope'] == 'os-neutron-net-id':
                network_id = tag['tag']
                break
        if not network_id:
            return
        metadata = []
        ctx = context.get_admin_context()
        port_filters = {'network_id': [network_id]}
        network_ports = plugin.get_ports(ctx, filters=port_filters)
        for port in network_ports:
            bindings = db.get_nsx_dhcp_bindings(ctx.session, port['id'])
            if bindings:
                # Should be only 1
                metadata.append({
                    'key': 'v4-static-binding%s' % bindings[0].nsx_binding_id,
                    'value': port['id'] + '-ipv4'})
        entry['metadata'] = metadata

    entries = get_resource_migration_data(
        nsxlib.logical_switch, [],
        'LOGICAL_SWITCH',
        resource_condition=cond,
        policy_resource_get=nsxpolicy.segment.get,
        policy_id_callback=get_policy_id,
        metadata_callback=add_metadata)
    migrate_resource(nsxlib, 'LOGICAL_SWITCH', entries,
                     MIGRATE_LIMIT_LOGICAL_SWITCH)


def migrate_ports(nsxlib, nsxpolicy, plugin):
    # For nsx networks support, keep a mapping of neutron id and MP id
    nsx_networks = {}
    ctx = context.get_admin_context()
    with ctx.session.begin(subtransactions=True):
        bindings = ctx.session.query(
            nsx_models.TzNetworkBinding).filter_by(
            binding_type='nsx-net').all()
        for bind in bindings:
            nsx_networks[bind.network_id] = bind.phy_uuid

    def get_policy_port(port_id, silent=False):
        # Get the segment id from neutron
        ctx = context.get_admin_context()
        neutron_port = plugin.get_port(ctx, port_id)
        net_id = neutron_port['network_id']
        if net_id in nsx_networks:
            segment_id = nsx_networks[net_id]
        else:
            segment_id = net_id
        return nsxpolicy.segment_port.get(segment_id, port_id, silent=silent)

    def add_metadata(entry, policy_id, resource):
        # Add binding maps with 'DEFAULT' key
        entry['metadata'] = [{'key': 'security-profile-binding-maps-id',
                              'value': policy_resources.DEFAULT_MAP_ID},
                             {'key': 'discovery-profile-binding-maps-id',
                              'value': policy_resources.DEFAULT_MAP_ID},
                             {'key': 'qos-profile-binding-maps-id',
                              'value': policy_resources.DEFAULT_MAP_ID}]

    entries = get_resource_migration_data(
        nsxlib.logical_port, ['os-neutron-port-id'],
        'LOGICAL_PORT',
        policy_resource_get=get_policy_port,
        metadata_callback=add_metadata)
    migrate_resource(nsxlib, 'LOGICAL_PORT', entries,
                     MIGRATE_LIMIT_LOGICAL_PORT)


def migrate_routers(nsxlib, nsxpolicy):

    entries = get_resource_migration_data(
        nsxlib.logical_router,
        ['os-neutron-router-id'],
        'TIER1',
        policy_resource_get=nsxpolicy.tier1.get,
        nsxlib_list_args={'router_type': nsx_constants.ROUTER_TYPE_TIER1})
    migrate_resource(nsxlib, 'TIER1', entries, MIGRATE_LIMIT_TIER1)
    migrated_routers = [entry['manager_id'] for entry in entries]
    return migrated_routers


def _get_subnet_by_cidr(subnets, cidr):
    for subnet in subnets:
        if subnet['cidr'] == cidr:
            return subnet['id']


def migrate_routers_config(nsxlib, nsxpolicy, plugin, migrated_routers):
    """Migrate advanced configuration of neutron Tier-1s
    This will use the list of Tier-1s migrated earlier
    """
    # Migrate all the centralized router ports and static routes for tier1
    # routers without specifying ids
    def get_policy_id(resource, policy_id):
        # No policy id needed here
        return

    def cond(resource):
        # Import ports only for the routers that were currently migrated
        # because there is no easy way to verify what was already migrated
        return resource['id'] in migrated_routers

    def add_metadata(entry, policy_id, resource):
        # Add router interfaces Ids
        ctx = context.get_admin_context()
        metadata = []
        mp_rtr_id = resource['id']
        router_ports = nsxlib.logical_router_port.get_by_router_id(mp_rtr_id)
        for port in router_ports:
            if 'linked_logical_switch_port_id' in port:
                lsp_id = port['linked_logical_switch_port_id']['target_id']
                lsp = nsxlib.logical_port.get(lsp_id)
                ls_id = lsp['logical_switch_id']
                if ls_id:
                    neutron_net_ids = plugin._get_neutron_net_ids_by_nsx_id(
                        ctx, ls_id)
                    if neutron_net_ids:
                        # Should be only 1
                        metadata.append({'key': port['id'],
                                         'value': neutron_net_ids[0]})
        # Add static routes ids
        static_routes = nsxlib.logical_router.list_static_routes(
            mp_rtr_id)['results']
        for route in static_routes:
            policy_id = "%s-%s" % (route['network'].replace('/', '_'),
                                   route['next_hops'][0]['ip_address'])
            metadata.append({'key': route['id'],
                             'value': policy_id})

        # Add locale-service id as <routerid>-0
        policy_id = None
        for tag in resource.get('tags', []):
            if tag['scope'] == 'os-neutron-router-id':
                policy_id = tag['tag']
        if policy_id:
            metadata.append({'key': 'localeServiceId',
                             'value': "%s-0" % policy_id})

        entry['metadata'] = metadata

    entries = get_resource_migration_data(
        nsxlib.logical_router,
        ['os-neutron-router-id'],
        'TIER1_LOGICAL_ROUTER_PORT',
        policy_id_callback=get_policy_id,
        resource_condition=cond,
        metadata_callback=add_metadata,
        skip_policy_path_check=True,
        nsxlib_list_args={'router_type': nsx_constants.ROUTER_TYPE_TIER1})
    migrate_resource(nsxlib, 'TIER1_LOGICAL_ROUTER_PORT', entries,
                     MIGRATE_LIMIT_TIER1_PORTS)

    # Migrate NAT rules per neutron tier1
    entries = []
    tier1s = nsxlib.logical_router.list(
        router_type=nsx_constants.ROUTER_TYPE_TIER1)['results']
    ctx = context.get_admin_context()
    for tier1 in tier1s:
        # skip routers that were not migrated in this script call
        tier1_mp_id = tier1['id']
        if tier1_mp_id not in migrated_routers:
            continue
        # skip non-neutron routers
        tier1_neutron_id = None
        for tag in tier1.get('tags', []):
            if tag['scope'] == 'os-neutron-router-id':
                tier1_neutron_id = tag['tag']
                break
        if not tier1_neutron_id:
            continue
        # Migrate each existing NAT rule, with the parameters the policy
        # plugin would have set
        router_subnets = plugin._load_router_subnet_cidrs_from_db(
            ctx, tier1_neutron_id)
        nat_rules = nsxlib.logical_router.list_nat_rules(
            tier1_mp_id)['results']
        for rule in nat_rules:
            # NO_DNAT rules for subnets
            if rule['action'] == 'NO_DNAT':
                seq_num = p_plugin.NAT_RULE_PRIORITY_GW
                cidr = rule['match_destination_network']
                subnet_id = _get_subnet_by_cidr(router_subnets, cidr)
                if not subnet_id:
                    LOG.error("Could not find subnet with cidr %s matching "
                              "NO_DNAT rule %s tier1 %s",
                              cidr, rule['id'], tier1_neutron_id)
                    continue
                policy_id = 'ND-' + subnet_id
            # SNAT rules for subnet or fip
            elif rule['action'] == 'SNAT':
                cidr = rule['match_source_network']
                if '/' in cidr:
                    seq_num = p_plugin.NAT_RULE_PRIORITY_GW
                    subnet_id = _get_subnet_by_cidr(router_subnets, cidr)
                    if not subnet_id:
                        LOG.error("Could not find subnet with cidr %s "
                                  "matching SNAT rule %s tier1 %s",
                                  cidr, rule['id'], tier1_neutron_id)
                        continue
                    policy_id = 'S-' + subnet_id
                else:
                    # FIP rule
                    seq_num = p_plugin.NAT_RULE_PRIORITY_FIP
                    fip_ip = rule['translated_network']
                    filters = {'floating_ip_address': [fip_ip]}
                    fips = plugin.get_floatingips(ctx, filters)
                    if not fips:
                        LOG.error("Could not find FIP with ip %s matching "
                                  "SNAT rule %s tier1 %s",
                                  fip_ip, rule['id'], tier1_neutron_id)
                        continue
                    policy_id = 'S-' + fips[0]['id']
            # DNAT rules for fip
            elif rule['action'] == 'DNAT':
                # FIP rule
                seq_num = p_plugin.NAT_RULE_PRIORITY_FIP
                fip_ip = rule['match_destination_network']
                filters = {'floating_ip_address': [fip_ip]}
                fips = plugin.get_floatingips(ctx, filters)
                if not fips:
                    LOG.error("Could not find FIP with ip %s matching DNAT "
                              "rule %s tier1 %s",
                              fip_ip, rule['id'], tier1_neutron_id)
                    continue
                policy_id = 'D-' + fips[0]['id']
            else:
                LOG.error("Unknown NAT action %s for rule %s tier1 %s",
                          rule['action'], rule['id'], tier1_neutron_id)
                continue

            entry = {'manager_id': rule['id'],
                     'policy_id': policy_id,
                     'metadata': [{'key': 'SEQUENCE_NUMBER',
                                   'value': seq_num}],
                     'linked_ids': [{'key': 'TIER1',
                                     'value': tier1_mp_id}]}
            entries.append(entry)
    migrate_resource(nsxlib, 'NAT', entries,
                     MIGRATE_LIMIT_NAT)


def migrate_tier0_config(nsxlib, nsxpolicy, tier0s):
    """Migrate ports and config for the already migrated Tier0s"""

    entries = []
    for tier0 in tier0s:
        uplink_ports = nsxlib.logical_router_port.get_tier0_uplink_ports(tier0)
        for uplink_port in uplink_ports:
            entries.append({'manager_id': uplink_port['id']})

    migrate_resource(nsxlib, 'TIER0_LOGICAL_ROUTER_PORT', entries,
                     MIGRATE_LIMIT_TIER0_PORTS, use_admin=True)

    def get_policy_id(resource, policy_id):
        # No policy id needed here
        return

    def cond(resource):
        # Import config only for the routers that were currently migrated
        # because there is no easy way to verify what was already migrated
        return resource['id'] in tier0s

    entries = get_resource_migration_data(
        nsxlib.logical_router, [],
        'TIER0_LOGICAL_ROUTER_CONFIG',
        policy_id_callback=get_policy_id,
        resource_condition=cond,
        skip_policy_path_check=True,
        nsxlib_list_args={'router_type': nsx_constants.ROUTER_TYPE_TIER0})
    migrate_resource(nsxlib, 'TIER0_LOGICAL_ROUTER_CONFIG', entries,
                     MIGRATE_LIMIT_TIER0, use_admin=True)


def migrate_groups(nsxlib, nsxpolicy):
    """Migrate NS groups of neutron defined security groups and predefined at
    plugin init
    """
    def get_policy_id_callback(res, policy_id):
        # In case of plugin init groups: give it the id the policy plugin
        # will use
        if res.get('display_name') == \
            v3_plugin.NSX_V3_FW_DEFAULT_NS_GROUP:
            return p_plugin.NSX_P_DEFAULT_GROUP

        if res.get('display_name') == \
            v3_plugin.NSX_V3_EXCLUDED_PORT_NSGROUP_NAME:
            return p_plugin.NSX_P_EXCLUDE_LIST_GROUP

        return policy_id

    def get_policy_group(group_id, silent=False):
        return nsxpolicy.group.get(policy_constants.DEFAULT_DOMAIN, group_id,
                                   silent=silent)

    entries = get_resource_migration_data(
        nsxlib.ns_group,
        ['os-neutron-secgr-id', 'os-neutron-id'],
        'NS_GROUP',
        policy_resource_get=get_policy_group,
        policy_id_callback=get_policy_id_callback)
    migrate_resource(nsxlib, 'NS_GROUP', entries, MIGRATE_LIMIT_NS_GROUP)


def dfw_migration_cond(resource):
    return (resource.get('enforced_on') == 'VIF' and
            resource.get('category') == 'Default' and
            resource.get('section_type') == 'LAYER3' and
            not resource.get('is_default') and
            # Migrate only DFW sections only and no edge FW sections
            'applied_tos' in resource and
            resource['applied_tos'][0].get('target_type', '') == 'NSGroup')


def migrate_dfw_sections(nsxlib, nsxpolicy, plugin):
    def get_policy_id_callback(res, policy_id):
        # In case of plugin init section: give it the id the policy plugin
        # will use
        if res.get('display_name') == \
            v3_plugin.NSX_V3_FW_DEFAULT_SECTION:
            return p_plugin.NSX_P_DEFAULT_SECTION

        return policy_id

    def add_metadata(entry, policy_id, resource):
        # Add category, sequence, domain, and rule ids
        ctx = context.get_admin_context()
        category = p_plugin.NSX_P_REGULAR_SECTION_CATEGORY
        if policy_id == p_plugin.NSX_P_DEFAULT_SECTION:
            category = p_plugin.NSX_P_DEFAULT_SECTION_CATEGORY
        else:
            try:
                sg = plugin.get_security_group(ctx, policy_id)
            except ext_sg.SecurityGroupNotFound:
                LOG.warning("Neutron SG %s was not found. Section %s may be "
                            "an orphaned", policy_id, resource['id'])
                provider = False
            else:
                provider = sg.get('provider')
            if provider:
                category = p_plugin.NSX_P_PROVIDER_SECTION_CATEGORY

        global DFW_SEQ
        metadata = [{'key': "category", 'value': category},
                    {'key': "sequence", 'value': str(DFW_SEQ)}]
        DFW_SEQ = DFW_SEQ + 1

        # Add the rules
        rules = nsxlib.firewall_section.get_rules(resource['id'])['results']
        linked_ids = []
        seq = 1
        for rule in rules:
            linked_ids.append({'key': rule['id'], 'value': str(seq)})
            if policy_id == p_plugin.NSX_P_DEFAULT_SECTION:
                # Default section rule ids are their seq numbers
                linked_ids.append({'key': "%s-policyid" % rule['id'],
                                   'value': seq})
            else:
                # The display name of the MP rule is the neutron id, and this
                # will become the policy id
                linked_ids.append({'key': "%s-policyid" % rule['id'],
                                   'value': rule['display_name']})
            seq = seq + 1
        entry['metadata'] = metadata
        entry['linked_ids'] = linked_ids

    def get_policy_section(sec_id, silent=False):
        return nsxpolicy.comm_map.get(policy_constants.DEFAULT_DOMAIN, sec_id,
                                      silent=silent)

    entries = get_resource_migration_data(
        nsxlib.firewall_section,
        ['os-neutron-secgr-id', 'os-neutron-id'],
        'DFW_SECTION', resource_condition=dfw_migration_cond,
        policy_resource_get=get_policy_section,
        policy_id_callback=get_policy_id_callback,
        metadata_callback=add_metadata)
    migrate_resource(nsxlib, 'DFW_SECTION', entries,
                     MIGRATE_LIMIT_DFW_SECTION,
                     count_internals=False)


def migrate_edge_firewalls(nsxlib, nsxpolicy, plugin):
    # -- Migrate edge firewall sections:
    # The MP plugin uses the default MP edge firewall section, while the policy
    # plugin uses a non default one, so regular migration cannot be used.
    # Instead, create new edge firewall sections, and remove rules from the MP
    # default sections

    # This is a hack to use the v3 plugin with the policy fwaas driver
    class MigrationNsxpFwaasCallbacks(fwaas_callbacks_v2.NsxpFwaasCallbacksV2):
        def __init__(self, with_rpc):
            super(MigrationNsxpFwaasCallbacks, self).__init__(with_rpc)
            # Make sure fwaas is considered as enabled
            self.fwaas_enabled = True

        def _get_port_firewall_group_id(self, ctx, port_id):
            # Override this api because directory.get_plugin does not work from
            # admin utils context.
            driver_db = firewall_db_v2.FirewallPluginDb()
            return driver_db.get_fwg_attached_to_port(ctx, port_id)

    fwaas_callbacks = MigrationNsxpFwaasCallbacks(False)
    plugin.nsxpolicy = nsxpolicy
    ctx = context.get_admin_context()
    routers = plugin.get_routers(ctx)
    global NSX_ROUTER_SECTIONS
    for rtr in routers:
        nsx_router_id = db.get_nsx_router_id(ctx.session, rtr['id'])
        nsx_rtr = nsxlib.logical_router.get(nsx_router_id)
        for sec in nsx_rtr.get('firewall_sections', []):
            section_id = sec['target_id']
            section = nsxlib.firewall_section.get(section_id)
            if section['display_name'] != 'Default LR Layer3 Section':
                continue
            rules = nsxlib.firewall_section.get_rules(section_id)['results']
            if len(rules) <= 1:
                continue
            # Non default rules exist. need to migrate this section
            router_db = plugin._get_router(ctx, rtr['id'])
            ports = plugin._get_router_interfaces(ctx, rtr['id'])
            fwaas_callbacks.update_router_firewall(
                ctx, rtr['id'], router_db, ports)
            LOG.debug("Created GW policy for router %s", rtr['id'])

            # delete rule from the default mp section at the end of the loop
            # so the new section will have time to realize
            NSX_ROUTER_SECTIONS.append({'id': section_id,
                                        'default_rule': rules[-1],
                                        'router_id': rtr['id']})


def migrate_dhcp_servers(nsxlib, nsxpolicy):
    # Each MP DHCP server will be migrated to a policy DHCP server config
    # which will be used by a segment later. It will get the neutron network id
    entries = get_resource_migration_data(
        nsxlib.dhcp_server,
        ['os-neutron-net-id'],
        'DHCP_SERVER',
        policy_resource_get=nsxpolicy.dhcp_server_config.get)
    migrate_resource(nsxlib, 'DHCP_SERVER', entries,
                     MIGRATE_LIMIT_DHCP_SERVER)


def migrate_lb_resources(nsxlib, nsxpolicy):
    migrate_lb_certificates(nsxlib, nsxpolicy)
    migrate_lb_monitors(nsxlib, nsxpolicy)
    migrate_lb_pools(nsxlib, nsxpolicy)
    migrate_lb_profiles(nsxlib, nsxpolicy)
    migrate_lb_listeners(nsxlib, nsxpolicy)
    migrate_lb_services(nsxlib, nsxpolicy)


def migrate_lb_certificates(nsxlib, nsxpolicy):
    entries = get_resource_migration_data(
        nsxlib.trust_management,
        [lb_const.LB_LISTENER_TYPE],
        'CERTIFICATE',
        policy_resource_get=nsxpolicy.certificate.get)
    migrate_resource(nsxlib, 'CERTIFICATE', entries,
                     MIGRATE_LIMIT_CERT)


def _migrate_lb_resource(nsxlib, nsxpolicy, neutron_tag, api_name,
                         migration_name, limit,
                         policy_api_name=None,
                         policy_id_callback=None):
    if not policy_api_name:
        policy_api_name = api_name
    entries = get_resource_migration_data(
        getattr(nsxlib.load_balancer, api_name),
        [neutron_tag],
        migration_name,
        policy_resource_get=getattr(nsxpolicy.load_balancer,
                                    policy_api_name).get,
        policy_id_callback=policy_id_callback)
    migrate_resource(nsxlib, migration_name, entries, limit)


def migrate_lb_listeners(nsxlib, nsxpolicy):
    _migrate_lb_resource(nsxlib, nsxpolicy,
                         lb_const.LB_LISTENER_TYPE,
                         'virtual_server',
                         'LB_VIRTUAL_SERVER',
                         MIGRATE_LIMIT_LB_VIRTUAL_SERVER)


def migrate_lb_pools(nsxlib, nsxpolicy):
    _migrate_lb_resource(nsxlib, nsxpolicy,
                         lb_const.LB_POOL_TYPE,
                         'pool',
                         'LB_POOL',
                         MIGRATE_LIMIT_LB_POOL,
                         policy_api_name='lb_pool')


def migrate_lb_monitors(nsxlib, nsxpolicy):
    _migrate_lb_resource(nsxlib, nsxpolicy,
                         lb_const.LB_HM_TYPE,
                         'monitor',
                         'LB_MONITOR',
                         MIGRATE_LIMIT_LB_MONITOR,
                         policy_api_name='lb_monitor_profile_http')


def migrate_lb_profiles(nsxlib, nsxpolicy):
    _migrate_lb_resource(nsxlib, nsxpolicy,
                         lb_const.LB_LISTENER_TYPE,
                         'application_profile',
                         'LB_APPLICATION_PROFILE',
                         MIGRATE_LIMIT_LB_APP_PROFILE,
                         policy_api_name='lb_http_profile')

    def get_policy_id_callback(res, policy_id):
        # The input policy id is the pool id
        # Need to add a suffix regarding the type of persistence
        if (res.get('resource_type') ==
            nsxlib_lb.PersistenceProfileTypes.SOURCE_IP):
            return "%s_%s" % (policy_id, 'sourceip')
        return "%s_%s" % (policy_id, 'cookie')

    _migrate_lb_resource(nsxlib, nsxpolicy,
                         lb_const.LB_POOL_TYPE,
                         'persistence_profile',
                         'LB_PERSISTENCE_PROFILE',
                         MIGRATE_LIMIT_LB_PER_PROFILE,
                         policy_api_name='lb_persistence_profile',
                         policy_id_callback=get_policy_id_callback)


def migrate_lb_services(nsxlib, nsxpolicy):

    def get_policy_id_callback(res, policy_id):
        # LB service is shared between few octavia loadbalancers
        # so the policy id is not the LB id, and those should be  marked
        # in the tags of the policy resource.
        # Keep the same id as MP so later we can search the MP DB
        # and update the tags
        return res['id']

    entries = get_resource_migration_data(
        nsxlib.load_balancer.service,
        ['os-api-version'],
        'LB_SERVICE',
        policy_resource_get=nsxpolicy.load_balancer.lb_service.get,
        policy_id_callback=get_policy_id_callback)
    migrate_resource(nsxlib, 'LB_SERVICE', entries,
                     MIGRATE_LIMIT_LB_SERVICE)


def migrate_t_resources_2_p(nsxlib, nsxpolicy, plugin):
    """Create policy resources for all MP resources used by neutron"""

    # Initialize the migration process
    if not ensure_migration_state_ready(nsxlib, with_abort=True):
        return False

    try:
        LOG.info("Starting resources migration")

        start_migration_process(nsxlib)

        # Migration order derives from the dependencies between resources
        public_switches, tier0s = migrate_tier0s(nsxlib, nsxpolicy, plugin)
        migrate_md_proxies(nsxlib, nsxpolicy, plugin)
        migrate_switch_profiles(nsxlib, nsxpolicy, plugin)
        migrate_groups(nsxlib, nsxpolicy)
        migrate_dhcp_servers(nsxlib, nsxpolicy)
        mp_routers = migrate_routers(nsxlib, nsxpolicy)
        migrate_networks(nsxlib, nsxpolicy, plugin, public_switches)
        migrate_ports(nsxlib, nsxpolicy, plugin)
        migrate_routers_config(nsxlib, nsxpolicy, plugin, mp_routers)
        migrate_tier0_config(nsxlib, nsxpolicy, tier0s)
        migrate_lb_resources(nsxlib, nsxpolicy)

        # Migrate firewall sections last as those take the longest to rollback
        # in case of error
        migrate_dfw_sections(nsxlib, nsxpolicy, plugin)
        migrate_edge_firewalls(nsxlib, nsxpolicy, plugin)

        # Finalize the migration (cause policy realization)
        end_migration_process(nsxlib)

        # Stop the migration service
        change_migration_service_status(start=False)

        return True

    except Exception as e:
        # Migration failed - abort it
        LOG.error("Exception occurred while making the request: %s", e)
        try:
            LOG.info("Aborting the current request")
            try:
                send_migration_plan_action(nsxlib, 'abort')
            except Exception as e:
                LOG.error("Abort migration failed: %s", e)

            global ROLLBACK_DATA
            if ROLLBACK_DATA:
                LOG.info("Rolling migration back %s", ROLLBACK_DATA)
                send_rollback_request(nsxlib,
                                      {'migration_data': ROLLBACK_DATA})
            # Finalize the migration (Also needed after rollback)
            end_migration_process(nsxlib)
            # Stop the migration service
            change_migration_service_status(start=False)
        except Exception as e:
            LOG.error("Rollback failed: %s", e)
        return False


def _get_network_nsx_segment_id(ctx, net_id):
    bindings = db.get_network_bindings(ctx.session, net_id)
    if (bindings and
        bindings[0].binding_type ==
        nsx_utils.NsxV3NetworkTypes.NSX_NETWORK):
        # return the ID of the NSX network
        return bindings[0].phy_uuid
    return net_id


def _delete_segment_profiles_bindings(nsxpolicy, segment_id):
    found = False
    sec_profiles = nsxpolicy.segment_security_profile_maps.list(segment_id)
    for profile in sec_profiles:
        found = True
        nsxpolicy.segment_security_profile_maps.delete(
            segment_id, profile['id'])

    qos_profiles = nsxpolicy.segment_qos_profile_maps.list(segment_id)
    for profile in qos_profiles:
        found = True
        nsxpolicy.segment_qos_profile_maps.delete(
            segment_id, profile['id'])

    discovery_profiles = nsxpolicy.segment_discovery_profile_maps.list(
        segment_id)
    for profile in discovery_profiles:
        found = True
        nsxpolicy.segment_discovery_profile_maps.delete(
            segment_id, profile['id'])

    if found:
        LOG.debug("Removed profiles mappings from segment %s", segment_id)


def post_migration_actions(nsxlib, nsxpolicy, nsxpolicy_admin, plugin):
    """Update created policy resources that does not match the policy plugins'
    expectations.
    """
    LOG.info("Starting post-migration actions")
    ctx = context.get_admin_context()

    # -- Update Lb tags on loadbalancer service
    pol_lb_services = nsxpolicy.load_balancer.lb_service.list()
    for lb_srv in pol_lb_services:
        # Verify this is a neutron resource
        if not is_neutron_resource(lb_srv):
            continue
        # Check if it already has the LB id tag
        migrated = False
        for tag in lb_srv.get('tags', []):
            if tag['scope'] == lb_utils.SERVICE_LB_TAG_SCOPE:
                migrated = True
                break
        if migrated:
            continue

        # Find the loadbalancers using this service from the DB
        lb_mapping = db.get_nsx_lbaas_loadbalancer_binding_by_service(
            ctx.session, lb_srv['id'])
        if lb_mapping:
            if 'tags' not in lb_srv:
                lb_srv['tags'] = []
            loadbalancers = [lb_map.loadbalancer_id for lb_map in lb_mapping]
            for lb_id in loadbalancers:
                lb_srv['tags'].append({'scope': lb_utils.SERVICE_LB_TAG_SCOPE,
                                       'tag': lb_id})
            nsxpolicy.load_balancer.lb_service.update(
                lb_srv['id'], tags=lb_srv['tags'])
            LOG.debug("Added tags to LB service %s", lb_srv['id'])

    # -- Update Lb L7 rules names
    mp_lb_rules = nsxlib.load_balancer.rule.list()['results']
    for mp_rule in mp_lb_rules:
        l7pol_id = None
        listener_id = None
        for tag in mp_rule.get('tags', []):
            if tag['scope'] == lb_const.LB_L7POLICY_TYPE:
                l7pol_id = tag['tag']
            if tag['scope'] == 'policyPath':
                listener_id = policy_utils.path_to_id(tag['tag'])

        if not l7pol_id or not listener_id:
            continue
        pol_vs = nsxpolicy.load_balancer.virtual_server.get(listener_id)
        pol_rules = pol_vs['rules']
        for pol_rule in pol_rules:
            if pol_rule['display_name'] == mp_rule['id']:
                new_name = nsx_utils.get_name_and_uuid('policy', l7pol_id)
                pol_rule['display_name'] = new_name
                nsxpolicy.load_balancer.virtual_server.update_lb_rules(
                    listener_id, pol_rules)
                LOG.debug("Updated L7 policy %s name on the virtual server",
                          l7pol_id)
                break

    # -- Create DHCP server configs to be used in neutron config
    # (The migration does not migrate MP DHCP profiles)
    neutron_dhcp = get_configured_values(plugin, '_native_dhcp_profile_uuid')
    for mp_dhcp in neutron_dhcp:
        # check if it was already migrated
        try:
            nsxpolicy.dhcp_server_config.get(mp_dhcp, silent=True)
        except Exception:
            # Create it
            mp_obj = nsxlib.native_dhcp_profile.get(mp_dhcp)
            # This should be created with the admin principal identity
            nsxpolicy_admin.dhcp_server_config.create_or_overwrite(
                mp_obj['display_name'],
                config_id=mp_dhcp,
                description=mp_obj.get('description', ''),
                edge_cluster_path=nsxpolicy.edge_cluster.get_path(
                    mp_obj['edge_cluster_id']))
            LOG.debug("Created DHCP server config %s for plugin config",
                      mp_dhcp)

    # -- Update Policy segments:
    # Set subnets GW for networks without linked routers
    # And remove unused segment profiles mappings
    networks = plugin.get_networks(ctx)
    for net in networks:
        if net.get('router:external'):
            continue
        seg_id = _get_network_nsx_segment_id(ctx, net['id'])
        if seg_id == net['id']:
            # This is not an nsx-net. Delete the bindings
            _delete_segment_profiles_bindings(nsxpolicy, seg_id)

        if plugin._get_network_router_ids(ctx, net['id']):
            continue
        # verify that this network has a dhcp subnet
        subnets = plugin.get_subnets_by_network(ctx, net['id'])
        for subnet in subnets:
            if subnet['ip_version'] == 4 and subnet['enable_dhcp']:
                # Update backend subnet
                segment = nsxpolicy.segment.get(seg_id)
                subnets = segment.get('subnets', [])
                if subnets and len(subnets) == 1 and subnet['gateway_ip']:
                    cidr_prefix = int(subnet['cidr'].split('/')[1])
                    gw = "%s/%s" % (subnet['gateway_ip'], cidr_prefix)
                    subnets[0]['gateway_address'] = gw
                nsxpolicy.segment.update(seg_id, subnets=subnets)
                LOG.debug("Updated gateway of network %s", net['id'])
                break

    # -- Delete MP edge firewall rules
    for section in NSX_ROUTER_SECTIONS:
        # make sure the policy section was already realized
        # with runtime_status=SUCESS
        nsxpolicy.gateway_policy.wait_until_state_sucessful(
            policy_constants.DEFAULT_DOMAIN, section['router_id'],
            max_attempts=600, with_refresh=True)
        nsxlib.firewall_section.update(
            section['id'], rules=[section['default_rule']])
        LOG.debug("Deleted MP edge FW section %s rules", section['id'])

    LOG.info("Post-migration actions done.")


def pre_migration_checks(nsxlib, plugin):
    """Check for unsupported configuration that will fail the migration"""
    nsx_version = nsxlib.get_version()
    if not nsx_utils.is_nsx_version_3_1_0(nsx_version):
        LOG.error("Pre migration check failed: Migration not supported for "
                  "NSX %s", nsx_version)
        return False

    # Cannot migrate with unsupported services
    service_plugins = cfg.CONF.service_plugins
    for srv_plugin in service_plugins:
        if 'vpnaas' in srv_plugin:
            LOG.error("Pre migration check failed: VPNaaS is not supported. "
                      "Please delete its configuration and disable it, before "
                      "running migration again.")
            return False
        if 'l2gw' in srv_plugin:
            LOG.error("Pre migration check failed: L2GW is not supported. "
                      "Please delete its configuration and disable it, before "
                      "running migration again.")
            return False

    # Tier0 with disabled BGP config
    neutron_t0s = get_neurton_tier0s(plugin)
    for tier0 in neutron_t0s:
        bgp_conf = nsxlib.logical_router.get_bgp_config(tier0)
        if not bgp_conf['enabled']:
            # Verify there are no neighbors configured
            if nsxlib.logical_router.get_bgp_neighbors(tier0)['result_count']:
                LOG.error("Pre migration check failed: Tier0 %s has BGP "
                          "neighbors but BGP is disabled. Please remove the "
                          "neighbors or enable BGP and try again.", tier0)
                return False

    # DHCP relay is unsupported
    if plugin._availability_zones_data.dhcp_relay_configured():
        LOG.error("Pre migration check failed: DHCP relay configuration "
                  "cannot be migrated. Please remove it from the plugin "
                  "configuration and from all NSX logical router ports and "
                  "try again.")
        return False

    return True


@admin_utils.output_header
def MP2Policy_pre_migration_check(resource, event, trigger, **kwargs):
    """Verify if the current configuration can be migrated to Policy"""
    nsxlib = utils.get_connected_nsxlib()
    with utils.NsxV3PluginWrapper() as plugin:
        if not pre_migration_checks(nsxlib, plugin):
            # Failed
            LOG.error("T2P migration cannot run. Please fix the configuration "
                      "and try again\n\n")
            sys.exit(1)


def _get_nsxlib_from_config(verbose):
    """Update the current config and return a working nsxlib
    or exit with error
    """

    if (not len(cfg.CONF.nsx_v3.nsx_api_user) or
        not len(cfg.CONF.nsx_v3.nsx_api_password)):
        LOG.error("T2P migration cannot run. Please provide nsx_api_user and "
                  "nsx_api_password in the configuration.")
        sys.exit(1)

    retriables = [nsxlib_exc.APITransactionAborted,
                  nsxlib_exc.ServerBusy]

    # Initialize the nsxlib objects, using just one of the managers because
    # the migration will be enabled only on one
    nsx_api_managers = copy.copy(cfg.CONF.nsx_v3.nsx_api_managers)
    nsx_api_user = copy.copy(cfg.CONF.nsx_v3.nsx_api_user)
    nsx_api_password = copy.copy(cfg.CONF.nsx_v3.nsx_api_password)

    for ind in range(len(nsx_api_managers)):
        # update the config to use this one manager only
        cfg.CONF.set_override(
            'nsx_api_managers', [nsx_api_managers[ind]], 'nsx_v3')
        if len(nsx_api_user) > ind:
            cfg.CONF.set_override(
                'nsx_api_user', [nsx_api_user[ind]], 'nsx_v3')
        else:
            cfg.CONF.set_override(
                'nsx_api_user', [nsx_api_user[0]], 'nsx_v3')
        if len(nsx_api_password) > ind:
            cfg.CONF.set_override(
                'nsx_api_password', [nsx_api_password[ind]], 'nsx_v3')
        else:
            cfg.CONF.set_override(
                'nsx_api_password', [nsx_api_password[0]], 'nsx_v3')
            utils.reset_global_nsxlib()
        nsxlib = utils.get_connected_nsxlib(verbose=verbose,
                                            allow_overwrite_header=True,
                                            retriable_exceptions=retriables)
        try:
            # test connectivity
            nsxlib.get_version()
        except Exception:
            LOG.warning("Failed to connect to NSX manager %s",
                        nsx_api_managers[ind])
        else:
            # Found a working manager
            return nsxlib

    LOG.error("T2P migration failed. Cannot connect to NSX with managers %s",
              nsx_api_managers)
    sys.exit(1)


@admin_utils.output_header
def MP2Policy_migration(resource, event, trigger, **kwargs):
    """Migrate NSX resources and neutron DB from NSX-T (MP) to Policy"""

    verbose = kwargs.get('verbose', False)
    if verbose:
        # Add DEBUG logs as well
        LOG.setLevel(logging.DEBUG)
    else:
        LOG.setLevel(logging.INFO)

    if kwargs.get('property'):
        # Add logfile
        properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
        logfile = properties.get('logfile', None)
        if logfile:
            f_handler = logging.FileHandler(logfile)
            f_formatter = logging.Formatter(
                '%(asctime)s %(levelname)s %(message)s')
            f_handler.setFormatter(f_formatter)
            LOG.addHandler(f_handler)

    nsxlib = _get_nsxlib_from_config(verbose)
    nsxpolicy = p_utils.get_connected_nsxpolicy(
        conf_path=cfg.CONF.nsx_v3, verbose=verbose)

    if cfg.CONF.nsx_v3.nsx_use_client_auth:
        # Also create a policy manager with admin user to manipulate
        # admin-defined resources which should not have neutron principal
        # identity
        nsxpolicy_admin = p_utils.get_connected_nsxpolicy(
            conf_path=cfg.CONF.nsx_v3,
            use_basic_auth=True,
            nsx_username=cfg.CONF.nsx_v3.nsx_api_user,
            nsx_password=cfg.CONF.nsx_v3.nsx_api_password,
            verbose=verbose)
    else:
        nsxpolicy_admin = nsxpolicy

    with utils.NsxV3PluginWrapper(verbose=verbose) as plugin:
        # Make sure FWaaS was initialized
        plugin.init_fwaas_for_admin_utils()

        start_time = time.time()
        if not pre_migration_checks(nsxlib, plugin):
            # Failed
            LOG.error("T2P migration cannot run. Please fix the configuration "
                      "and try again\n\n")
            sys.exit(1)
        elapsed_time = time.time() - start_time
        LOG.debug("Pre-migration took %s seconds", elapsed_time)

        start_time = time.time()
        if not migrate_t_resources_2_p(nsxlib, nsxpolicy, plugin):
            # Failed
            LOG.error("T2P migration failed. Aborting\n\n")
            sys.exit(1)
        elapsed_time = time.time() - start_time
        LOG.debug("Migration took %s seconds", elapsed_time)

        start_time = time.time()
        post_migration_actions(nsxlib, nsxpolicy, nsxpolicy_admin, plugin)
        elapsed_time = time.time() - start_time
        LOG.debug("Post-migration took %s seconds", elapsed_time)

    LOG.info("T2P migration completed successfully\n\n")


@admin_utils.output_header
def MP2Policy_cleanup_db_mappings(resource, event, trigger, **kwargs):
    """Delete all entries from nsx-t mapping tables in DB"""
    confirm = admin_utils.query_yes_no(
        "Are you sure you want to delete all MP plugin mapping DB tables?",
        default="no")
    if not confirm:
        LOG.info("Deletion aborted by user")
        return

    ctx = context.get_admin_context()
    mp_mapping_tables = [nsx_models.NeutronNsxFirewallSectionMapping,
                         nsx_models.NeutronNsxSecurityGroupMapping,
                         nsx_models.NeutronNsxRuleMapping,
                         nsx_models.NeutronNsxPortMapping,
                         nsx_models.NeutronNsxRouterMapping,
                         nsx_models.NeutronNsxServiceBinding,
                         nsx_models.NeutronNsxDhcpBinding,
                         nsx_models.QosPolicySwitchProfile,
                         nsx_models.NsxLbaasLoadbalancer,
                         nsx_models.NsxLbaasListener,
                         nsx_models.NsxLbaasPool,
                         nsx_models.NsxLbaasMonitor,
                         nsx_models.NsxLbaasL7Rule,
                         nsx_models.NsxLbaasL7Policy]
    for table in mp_mapping_tables:
        ctx.session.query(table).delete()

    LOG.info("Deleted all MP plugin mapping DB tables.")


registry.subscribe(MP2Policy_migration,
                   constants.NSX_MIGRATE_T_P,
                   shell.Operations.IMPORT.value)

registry.subscribe(MP2Policy_pre_migration_check,
                   constants.NSX_MIGRATE_T_P,
                   shell.Operations.VALIDATE.value)

registry.subscribe(MP2Policy_cleanup_db_mappings,
                   constants.NSX_MIGRATE_T_P,
                   shell.Operations.CLEAN_ALL.value)



# Copyright 2019 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db.models import servicetype as st_db
from neutron.db import models_v2
from neutron_lib.db import constants as db_const
from neutron_lib.db import model_base
import sqlalchemy as sa
from sqlalchemy.ext import orderinglist
from sqlalchemy import orm

from vmware_nsx.tests.unit.services.lbaas import lb_constants as lb_const


class SessionPersistenceV2(model_base.BASEV2):

    NAME = 'session_persistence'

    __tablename__ = "lbaas_sessionpersistences"

    pool_id = sa.Column(sa.String(36),
                        sa.ForeignKey("lbaas_pools.id"),
                        primary_key=True,
                        nullable=False)
    type = sa.Column(sa.Enum(*lb_const.SUPPORTED_SP_TYPES,
                             name="lbaas_sesssionpersistences_typev2"),
                     nullable=False)
    cookie_name = sa.Column(sa.String(1024), nullable=True)


class LoadBalancerStatistics(model_base.BASEV2):
    """Represents load balancer statistics."""

    NAME = 'loadbalancer_stats'

    __tablename__ = "lbaas_loadbalancer_statistics"

    loadbalancer_id = sa.Column(sa.String(36),
                                sa.ForeignKey("lbaas_loadbalancers.id"),
                                primary_key=True,
                                nullable=False)
    bytes_in = sa.Column(sa.BigInteger, nullable=False)
    bytes_out = sa.Column(sa.BigInteger, nullable=False)
    active_connections = sa.Column(sa.BigInteger, nullable=False)
    total_connections = sa.Column(sa.BigInteger, nullable=False)

    @orm.validates('bytes_in', 'bytes_out',
                   'active_connections', 'total_connections')
    def validate_non_negative_int(self, key, value):
        if value < 0:
            data = {'key': key, 'value': value}
            raise ValueError('The %(key)s field can not have '
                             'negative value. '
                             'Current value is %(value)d.' % data)
        return value


class MemberV2(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents a v2 neutron load balancer member."""

    NAME = 'member'

    __tablename__ = "lbaas_members"

    __table_args__ = (
        sa.schema.UniqueConstraint('pool_id', 'address', 'protocol_port',
                                   name='uniq_pool_address_port_v2'),
    )
    pool_id = sa.Column(sa.String(36), sa.ForeignKey("lbaas_pools.id"),
                        nullable=False)
    address = sa.Column(sa.String(64), nullable=False)
    protocol_port = sa.Column(sa.Integer, nullable=False)
    weight = sa.Column(sa.Integer, nullable=True)
    admin_state_up = sa.Column(sa.Boolean(), nullable=False)
    subnet_id = sa.Column(sa.String(36), nullable=True)
    provisioning_status = sa.Column(sa.String(16), nullable=False)
    operating_status = sa.Column(sa.String(16), nullable=False)
    name = sa.Column(sa.String(db_const.NAME_FIELD_SIZE), nullable=True)

    @property
    def root_loadbalancer(self):
        return self.pool.loadbalancer

    @property
    def to_api_dict(self):
        def to_dict(sa_model, attributes):
            ret = {}
            for attr in attributes:
                value = getattr(sa_model, attr)
                ret[attr] = value
            return ret

        ret_dict = to_dict(self, [
            'id', 'tenant_id', 'pool_id', 'address', 'protocol_port', 'weight',
            'admin_state_up', 'subnet_id', 'name'])

        return ret_dict


class HealthMonitorV2(model_base.BASEV2, model_base.HasId,
                      model_base.HasProject):
    """Represents a v2 neutron load balancer healthmonitor."""

    NAME = 'healthmonitor'

    __tablename__ = "lbaas_healthmonitors"

    type = sa.Column(sa.Enum(*lb_const.SUPPORTED_HEALTH_MONITOR_TYPES,
                             name="healthmonitors_typev2"),
                     nullable=False)
    delay = sa.Column(sa.Integer, nullable=False)
    timeout = sa.Column(sa.Integer, nullable=False)
    max_retries = sa.Column(sa.Integer, nullable=False)
    http_method = sa.Column(sa.String(16), nullable=True)
    url_path = sa.Column(sa.String(255), nullable=True)
    expected_codes = sa.Column(sa.String(64), nullable=True)
    provisioning_status = sa.Column(sa.String(16), nullable=False)
    admin_state_up = sa.Column(sa.Boolean(), nullable=False)
    name = sa.Column(sa.String(db_const.NAME_FIELD_SIZE), nullable=True)
    max_retries_down = sa.Column(sa.Integer, nullable=True)

    @property
    def root_loadbalancer(self):
        return self.pool.loadbalancer

    @property
    def to_api_dict(self):
        def to_dict(sa_model, attributes):
            ret = {}
            for attr in attributes:
                value = getattr(sa_model, attr)
                ret[attr] = value
            return ret

        ret_dict = to_dict(self, [
            'id', 'tenant_id', 'type', 'delay', 'timeout', 'max_retries',
            'http_method', 'url_path', 'expected_codes', 'admin_state_up',
            'name', 'max_retries_down'])

        ret_dict['pools'] = []
        if self.pool:
            ret_dict['pools'].append({'id': self.pool.id})
        if self.type in [lb_const.HEALTH_MONITOR_TCP,
                         lb_const.HEALTH_MONITOR_PING]:
            ret_dict.pop('http_method')
            ret_dict.pop('url_path')
            ret_dict.pop('expected_codes')

        return ret_dict


class LoadBalancer(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents a v2 neutron load balancer."""

    NAME = 'loadbalancer'

    __tablename__ = "lbaas_loadbalancers"

    name = sa.Column(sa.String(255))
    description = sa.Column(sa.String(255))
    vip_subnet_id = sa.Column(sa.String(36), nullable=False)
    vip_port_id = sa.Column(sa.String(36), sa.ForeignKey(
        'ports.id', name='fk_lbaas_loadbalancers_ports_id'))
    vip_address = sa.Column(sa.String(36))
    provisioning_status = sa.Column(sa.String(16), nullable=False)
    operating_status = sa.Column(sa.String(16), nullable=False)
    admin_state_up = sa.Column(sa.Boolean(), nullable=False)
    vip_port = orm.relationship(models_v2.Port)
    stats = orm.relationship(
        LoadBalancerStatistics,
        uselist=False,
        backref=orm.backref("loadbalancer", uselist=False),
        cascade="all, delete-orphan")
    provider = orm.relationship(
        st_db.ProviderResourceAssociation,
        uselist=False,
        primaryjoin="LoadBalancer.id==ProviderResourceAssociation.resource_id",
        foreign_keys=[st_db.ProviderResourceAssociation.resource_id],
        # NOTE(ihrachys) it's not exactly clear why we would need to have the
        # backref created (and not e.g. just back_populates= link) since we
        # don't use the reverse property anywhere, but it helps with
        # accommodating to the new neutron code that automatically detects
        # obsolete foreign key state and expires affected relationships. The
        # code is located in neutron/db/api.py and assumes all relationships
        # should have backrefs.
        backref='loadbalancer',
        # this is only for old API backwards compatibility because when a load
        # balancer is deleted the pool ID should be the same as the load
        # balancer ID and should not be cleared out in this table
        viewonly=True)
    flavor_id = sa.Column(sa.String(36), sa.ForeignKey(
        'flavors.id', name='fk_lbaas_loadbalancers_flavors_id'))

    @property
    def root_loadbalancer(self):
        return self

    @property
    def to_api_dict(self):
        def to_dict(sa_model, attributes):
            ret = {}
            for attr in attributes:
                value = getattr(sa_model, attr)
                ret[attr] = value
            return ret

        ret_dict = to_dict(self, [
            'id', 'tenant_id', 'name', 'description',
            'vip_subnet_id', 'vip_port_id', 'vip_address', 'operating_status',
            'provisioning_status', 'admin_state_up', 'flavor_id'])
        ret_dict['listeners'] = [{'id': listener.id}
                                 for listener in self.listeners]
        ret_dict['pools'] = [{'id': pool.id} for pool in self.pools]

        if self.provider:
            ret_dict['provider'] = self.provider.provider_name

        if not self.flavor_id:
            del ret_dict['flavor_id']

        return ret_dict


class PoolV2(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents a v2 neutron load balancer pool."""

    NAME = 'pool'

    __tablename__ = "lbaas_pools"

    name = sa.Column(sa.String(255), nullable=True)
    description = sa.Column(sa.String(255), nullable=True)
    loadbalancer_id = sa.Column(sa.String(36), sa.ForeignKey(
        "lbaas_loadbalancers.id"))
    healthmonitor_id = sa.Column(sa.String(36),
                                 sa.ForeignKey("lbaas_healthmonitors.id"),
                                 unique=True,
                                 nullable=True)
    protocol = sa.Column(sa.Enum(*lb_const.POOL_SUPPORTED_PROTOCOLS,
                                 name="pool_protocolsv2"),
                         nullable=False)
    lb_algorithm = sa.Column(sa.Enum(*lb_const.SUPPORTED_LB_ALGORITHMS,
                                     name="lb_algorithmsv2"),
                             nullable=False)
    admin_state_up = sa.Column(sa.Boolean(), nullable=False)
    provisioning_status = sa.Column(sa.String(16), nullable=False)
    operating_status = sa.Column(sa.String(16), nullable=False)
    members = orm.relationship(MemberV2,
                               backref=orm.backref("pool", uselist=False),
                               cascade="all, delete-orphan")
    healthmonitor = orm.relationship(
        HealthMonitorV2,
        backref=orm.backref("pool", uselist=False))
    session_persistence = orm.relationship(
        SessionPersistenceV2,
        uselist=False,
        backref=orm.backref("pool", uselist=False),
        cascade="all, delete-orphan")
    loadbalancer = orm.relationship(
        LoadBalancer, uselist=False,
        backref=orm.backref("pools", uselist=True))

    @property
    def root_loadbalancer(self):
        return self.loadbalancer

    # No real relationship here. But we want to fake a pool having a
    # 'listener_id' sometimes for API back-ward compatibility purposes.
    @property
    def listener(self):
        if self.listeners:
            return self.listeners[0]
        else:
            return None

    @property
    def to_api_dict(self):
        def to_dict(sa_model, attributes):
            ret = {}
            for attr in attributes:
                value = getattr(sa_model, attr)
                ret[attr] = value
            return ret

        ret_dict = to_dict(self, [
            'id', 'tenant_id', 'name', 'description',
            'healthmonitor_id', 'protocol', 'lb_algorithm', 'admin_state_up'])

        ret_dict['loadbalancers'] = []
        if self.loadbalancer:
            ret_dict['loadbalancers'].append({'id': self.loadbalancer.id})
        ret_dict['session_persistence'] = None
        if self.session_persistence:
            ret_dict['session_persistence'] = (
                to_dict(self.session_persistence, [
                    'type', 'cookie_name']))
        ret_dict['members'] = [{'id': member.id} for member in self.members]
        ret_dict['listeners'] = [{'id': listener.id}
                                 for listener in self.listeners]
        if self.listener:
            ret_dict['listener_id'] = self.listener.id
        else:
            ret_dict['listener_id'] = None
        ret_dict['l7_policies'] = [{'id': l7_policy.id}
            for l7_policy in self.l7_policies]
        return ret_dict


class SNI(model_base.BASEV2):

    """Many-to-many association between Listener and TLS container ids
    Making the SNI certificates list, ordered using the position
    """

    NAME = 'sni'

    __tablename__ = "lbaas_sni"

    listener_id = sa.Column(sa.String(36),
                            sa.ForeignKey("lbaas_listeners.id"),
                            primary_key=True,
                            nullable=False)
    tls_container_id = sa.Column(sa.String(128),
                                 primary_key=True,
                                 nullable=False)
    position = sa.Column(sa.Integer)

    @property
    def root_loadbalancer(self):
        return self.listener.loadbalancer


class L7Rule(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents L7 Rule."""

    NAME = 'l7rule'

    __tablename__ = "lbaas_l7rules"

    l7policy_id = sa.Column(sa.String(36),
                            sa.ForeignKey("lbaas_l7policies.id"),
                            nullable=False)
    type = sa.Column(sa.Enum(*lb_const.SUPPORTED_L7_RULE_TYPES,
                             name="l7rule_typesv2"),
                     nullable=False)
    compare_type = sa.Column(sa.Enum(*lb_const.SUPPORTED_L7_RULE_COMPARE_TYPES,
                                     name="l7rule_compare_typev2"),
                             nullable=False)
    invert = sa.Column(sa.Boolean(), nullable=False)
    key = sa.Column(sa.String(255), nullable=True)
    value = sa.Column(sa.String(255), nullable=False)
    provisioning_status = sa.Column(sa.String(16), nullable=False)
    admin_state_up = sa.Column(sa.Boolean(), nullable=False)

    @property
    def root_loadbalancer(self):
        return self.policy.listener.loadbalancer

    @property
    def to_api_dict(self):
        def to_dict(sa_model, attributes):
            ret = {}
            for attr in attributes:
                value = getattr(sa_model, attr)
                ret[attr] = value
            return ret

        ret_dict = to_dict(self, [
            'id', 'tenant_id', 'type', 'compare_type', 'invert', 'key',
            'value', 'admin_state_up'])

        ret_dict['policies'] = []
        if self.policy:
            ret_dict['policies'].append({'id': self.policy.id})
        return ret_dict


class L7Policy(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents L7 Policy."""

    NAME = 'l7policy'

    __tablename__ = "lbaas_l7policies"

    name = sa.Column(sa.String(255), nullable=True)
    description = sa.Column(sa.String(255), nullable=True)
    listener_id = sa.Column(sa.String(36),
                            sa.ForeignKey("lbaas_listeners.id"),
                            nullable=False)
    action = sa.Column(sa.Enum(*lb_const.SUPPORTED_L7_POLICY_ACTIONS,
                               name="l7policy_action_typesv2"),
                       nullable=False)
    redirect_pool_id = sa.Column(sa.String(36),
                                 sa.ForeignKey("lbaas_pools.id"),
                                 nullable=True)
    redirect_url = sa.Column(sa.String(255),
                             nullable=True)
    position = sa.Column(sa.Integer, nullable=False)
    provisioning_status = sa.Column(sa.String(16), nullable=False)
    admin_state_up = sa.Column(sa.Boolean(), nullable=False)
    rules = orm.relationship(
        L7Rule,
        uselist=True,
        primaryjoin="L7Policy.id==L7Rule.l7policy_id",
        foreign_keys=[L7Rule.l7policy_id],
        cascade="all, delete-orphan",
        backref=orm.backref("policy")
    )
    redirect_pool = orm.relationship(
        PoolV2, backref=orm.backref("l7_policies", uselist=True))

    @property
    def root_loadbalancer(self):
        return self.listener.loadbalancer

    @property
    def to_api_dict(self):
        def to_dict(sa_model, attributes):
            ret = {}
            for attr in attributes:
                value = getattr(sa_model, attr)
                ret[attr] = value
            return ret

        ret_dict = to_dict(self, [
            'id', 'tenant_id', 'name', 'description', 'listener_id', 'action',
            'redirect_pool_id', 'redirect_url', 'position', 'admin_state_up'])

        ret_dict['listeners'] = [{'id': self.listener_id}]
        ret_dict['rules'] = [{'id': rule.id} for rule in self.rules]
        if (ret_dict.get('action') ==
                lb_const.L7_POLICY_ACTION_REDIRECT_TO_POOL):
            del ret_dict['redirect_url']
        return ret_dict


class Listener(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents a v2 neutron listener."""

    NAME = 'listener'

    __tablename__ = "lbaas_listeners"

    __table_args__ = (
        sa.schema.UniqueConstraint('loadbalancer_id', 'protocol_port',
                                   name='uniq_loadbalancer_listener_port'),
    )

    name = sa.Column(sa.String(255))
    description = sa.Column(sa.String(255))
    default_pool_id = sa.Column(sa.String(36), sa.ForeignKey("lbaas_pools.id"),
                                nullable=True)
    loadbalancer_id = sa.Column(sa.String(36), sa.ForeignKey(
        "lbaas_loadbalancers.id"))
    protocol = sa.Column(sa.Enum(*lb_const.LISTENER_SUPPORTED_PROTOCOLS,
                                 name="listener_protocolsv2"),
                         nullable=False)
    default_tls_container_id = sa.Column(sa.String(128),
                                         default=None, nullable=True)
    sni_containers = orm.relationship(
            SNI,
            backref=orm.backref("listener", uselist=False),
            uselist=True,
            primaryjoin="Listener.id==SNI.listener_id",
            order_by='SNI.position',
            collection_class=orderinglist.ordering_list(
                'position'),
            foreign_keys=[SNI.listener_id],
            cascade="all, delete-orphan"
    )
    protocol_port = sa.Column(sa.Integer, nullable=False)
    connection_limit = sa.Column(sa.Integer)
    admin_state_up = sa.Column(sa.Boolean(), nullable=False)
    provisioning_status = sa.Column(sa.String(16), nullable=False)
    operating_status = sa.Column(sa.String(16), nullable=False)
    default_pool = orm.relationship(
        PoolV2, backref=orm.backref("listeners"))
    loadbalancer = orm.relationship(
        LoadBalancer,
        backref=orm.backref("listeners", uselist=True))
    l7_policies = orm.relationship(
        L7Policy,
        uselist=True,
        primaryjoin="Listener.id==L7Policy.listener_id",
        order_by="L7Policy.position",
        collection_class=orderinglist.ordering_list('position', count_from=1),
        foreign_keys=[L7Policy.listener_id],
        cascade="all, delete-orphan",
        backref=orm.backref("listener"))

    @property
    def root_loadbalancer(self):
        return self.loadbalancer

    @property
    def to_api_dict(self):
        def to_dict(sa_model, attributes):
            ret = {}
            for attr in attributes:
                value = getattr(sa_model, attr)
                ret[attr] = value
            return ret

        ret_dict = to_dict(self, [
            'id', 'tenant_id', 'name', 'description', 'default_pool_id',
            'protocol', 'default_tls_container_id', 'protocol_port',
            'connection_limit', 'admin_state_up'])

        # NOTE(blogan): Returning a list to future proof for M:N objects
        # that are not yet implemented.
        ret_dict['loadbalancers'] = []
        if self.loadbalancer:
            ret_dict['loadbalancers'].append({'id': self.loadbalancer.id})
        ret_dict['sni_container_refs'] = [container.tls_container_id
                                          for container in self.sni_containers]
        ret_dict['default_tls_container_ref'] = self.default_tls_container_id
        ret_dict['l7policies'] = [{'id': l7_policy.id}
            for l7_policy in self.l7_policies]
        return ret_dict



# Copyright 2016 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_config import cfg
from oslo_log import log as logging
from sqlalchemy.orm import exc

from neutron.db.models import securitygroup
from neutron_lib.api.definitions import allowedaddresspairs as addr_apidef
from neutron_lib import constants as const

from vmware_nsx.db import db as nsx_db
from vmware_nsx.db import nsx_models
from vmware_nsx.services.qos.common import utils as qos_utils
from vmware_nsxlib.v3 import core_resources
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3 import nsx_constants

NSX_V3_PSEC_PROFILE_NAME = 'neutron_port_spoof_guard_profile'
NSX_V3_DHCP_PROFILE_NAME = 'neutron_port_dhcp_profile'

PORT_ERROR_TYPE_MISSING = "Missing port"
PORT_ERROR_TYPE_PROFILE = "Wrong switching profiles"
PORT_ERROR_TYPE_BINDINGS = "Wrong address binding"

# Default UUID for the global OS rule
NSX_V3_OS_DFW_UUID = '00000000-def0-0000-0fed-000000000000'

LOG = logging.getLogger(__name__)


def get_orphaned_dhcp_servers(context, plugin, nsxlib, dhcp_profile_uuid=None):
    # An orphaned DHCP server means the associated neutron network
    # does not exist or has no DHCP-enabled subnet.

    orphaned_servers = []
    server_net_pairs = []

    # Find matching DHCP servers (for a given dhcp_profile_uuid).
    response = nsxlib.dhcp_server.list()
    for dhcp_server in response['results']:
        if (dhcp_profile_uuid and
            dhcp_server['dhcp_profile_id'] != dhcp_profile_uuid):
            continue
        found = False
        neutron_obj = False
        for tag in dhcp_server.get('tags', []):
            if tag['scope'] == 'os-neutron-net-id':
                dhcp_server['neutron_net_id'] = tag['tag']
                server_net_pairs.append((dhcp_server, tag['tag']))
                found = True
            if tag['scope'] == 'os-api-version':
                neutron_obj = True
        if not found and neutron_obj:
            # The associated neutron network is not defined.
            dhcp_server['neutron_net_id'] = None
            orphaned_servers.append(dhcp_server)

    # Check if there is DHCP-enabled subnet in each network.
    for dhcp_server, net_id in server_net_pairs:
        try:
            network = plugin.get_network(context, net_id)
        except Exception:
            # The associated neutron network is not found in DB.
            orphaned_servers.append(dhcp_server)
            continue
        dhcp_enabled = False
        for subnet_id in network['subnets']:
            subnet = plugin.get_subnet(context, subnet_id)
            if subnet['enable_dhcp']:
                dhcp_enabled = True
                break
        if not dhcp_enabled:
            orphaned_servers.append(dhcp_server)

    return orphaned_servers


def delete_orphaned_dhcp_server(context, nsxlib, server):
    # Delete an orphaned DHCP server:
    # (1) delete the attached logical DHCP port,
    # (2) delete the logical DHCP server,
    # (3) clean corresponding neutron DB entry.
    # Return True if it was deleted, or false + error if not
    try:
        response = nsxlib.logical_port.get_by_attachment('DHCP_SERVICE',
                                                         server['id'])
        if response and response['result_count'] > 0:
            nsxlib.logical_port.delete(response['results'][0]['id'])
        nsxlib.dhcp_server.delete(server['id'])
        net_id = server.get('neutron_net_id')
        if net_id:
            # Delete neutron_net_id -> dhcp_service_id mapping from the DB.
            nsx_db.delete_neutron_nsx_service_binding(
                context.session, net_id,
                nsx_constants.SERVICE_DHCP)
        return True, None
    except Exception as e:
        return False, e


def get_orphaned_networks(context, nsxlib):
    nsx_switches = nsxlib.logical_switch.list()['results']
    missing_networks = []
    for nsx_switch in nsx_switches:
        # check if it exists in the neutron DB
        net_ids = nsx_db.get_net_ids(context.session, nsx_switch['id'])
        if not net_ids:
            # Skip non-neutron networks, by tags
            neutron_net = False
            for tag in nsx_switch.get('tags', []):
                if tag.get('scope') == 'os-neutron-net-id':
                    neutron_net = True
                    nsx_switch['neutron_net_id'] = tag['tag']
                    break
            if neutron_net:
                missing_networks.append(nsx_switch)
    return missing_networks


def get_orphaned_routers(context, nsxlib):
    nsx_routers = nsxlib.logical_router.list()['results']
    missing_routers = []
    for nsx_router in nsx_routers:
        # check if it exists in the neutron DB
        neutron_id = nsx_db.get_neutron_from_nsx_router_id(context.session,
                                                           nsx_router['id'])
        if not neutron_id:
            # Skip non-neutron routers, by tags
            for tag in nsx_router.get('tags', []):
                if tag.get('scope') == 'os-neutron-router-id':
                    nsx_router['neutron_router_id'] = tag['tag']
                    missing_routers.append(nsx_router)
                    break
    return missing_routers


def delete_orphaned_router(nsxlib, nsx_id):
    # Delete an orphaned logical router from the NSX:
    # (1) delete the attached ports,
    # (2) delete the logical router
    # Return True if it was deleted, or false + error if not
    try:
        # first delete its ports
        ports = nsxlib.logical_router_port.get_by_router_id(nsx_id)
        for port in ports:
            nsxlib.logical_router_port.delete(port['id'])
        nsxlib.logical_router.delete(nsx_id)
    except Exception as e:
        return False, e
    else:
        return True, None


def get_security_groups_mappings(context):
    q = context.session.query(
        securitygroup.SecurityGroup.name,
        securitygroup.SecurityGroup.id,
        nsx_models.NeutronNsxFirewallSectionMapping.nsx_id,
        nsx_models.NeutronNsxSecurityGroupMapping.nsx_id).join(
            nsx_models.NeutronNsxFirewallSectionMapping,
            nsx_models.NeutronNsxSecurityGroupMapping).all()
    sg_mappings = [{'name': mapp[0],
                    'id': mapp[1],
                    'section-id': mapp[2],
                    'nsx-securitygroup-id': mapp[3]}
                   for mapp in q]
    return sg_mappings


def get_orphaned_firewall_sections(context, nsxlib):
    orphaned_sections = []
    fw_sections = nsxlib.firewall_section.list()
    if not fw_sections:
        return orphaned_sections
    sg_mappings = get_security_groups_mappings(context)
    for fw_section in fw_sections:
        for sg_db in sg_mappings:
            if fw_section['id'] == sg_db['section-id']:
                break
        else:
            # Skip non-neutron sections, by tags
            neutron_obj = False
            for tag in fw_section.get('tags', []):
                if tag['scope'] == 'os-api-version':
                    neutron_obj = True
                if tag.get('scope') == 'os-neutron-secgr-id':
                    fw_section['neutron_sg_id'] = tag['tag']
            if neutron_obj:
                orphaned_sections.append(fw_section)
    return orphaned_sections


def get_security_group_rules_mappings(context):
    q = context.session.query(
        securitygroup.SecurityGroupRule.id,
        nsx_models.NeutronNsxRuleMapping.nsx_id).join(
            nsx_models.NeutronNsxRuleMapping).all()
    sg_mappings = [{'rule_id': mapp[0],
                    'nsx_rule_id': mapp[1]}
                   for mapp in q]
    return sg_mappings


def get_orphaned_firewall_section_rules(context, nsxlib):
    fw_sections = nsxlib.firewall_section.list()
    sg_mappings = get_security_groups_mappings(context)
    rules_mappings = get_security_group_rules_mappings(context)
    orphaned_rules = []
    nsx_rules_in_mappings = [r['nsx_rule_id'] for r in rules_mappings]
    for fw_section in fw_sections:
        for sg_db in sg_mappings:
            if (fw_section['id'] == sg_db['section-id'] and
                sg_db['id'] != NSX_V3_OS_DFW_UUID):
                # found the right neutron SG
                section_rules = nsxlib.firewall_section.get_rules(
                    fw_section['id'])['results']
                for nsx_rule in section_rules:
                    if nsx_rule['id'] not in nsx_rules_in_mappings:
                        # orphaned rule
                        orphaned_rules.append(
                            {'security-group-name': sg_db['name'],
                             'security-group-id': sg_db['id'],
                             'section-id': fw_section['id'],
                             'rule-id': nsx_rule['id']})
    return orphaned_rules


def get_dhcp_profile_id(nsxlib):
    profiles = nsxlib.switching_profile.find_by_display_name(
        NSX_V3_DHCP_PROFILE_NAME)
    if profiles and len(profiles) == 1:
        return profiles[0]['id']
    LOG.warning("Could not find DHCP profile on backend")


def get_spoofguard_profile_id(nsxlib):
    profiles = nsxlib.switching_profile.find_by_display_name(
        NSX_V3_PSEC_PROFILE_NAME)
    if profiles and len(profiles) == 1:
        return profiles[0]['id']
    LOG.warning("Could not find Spoof Guard profile on backend")


def add_profile_mismatch(problems, neutron_id, nsx_id, prf_id, title):
    msg = ('Wrong %(title)s profile %(prf_id)s') % {'title': title,
                                                    'prf_id': prf_id}
    problems.append({'neutron_id': neutron_id,
                     'nsx_id': nsx_id,
                     'error': msg,
                     'error_type': PORT_ERROR_TYPE_PROFILE})


def get_port_nsx_id(session, neutron_id):
    # get the nsx port id from the DB mapping
    try:
        mapping = (session.query(nsx_models.NeutronNsxPortMapping).
                   filter_by(neutron_id=neutron_id).
                   one())
        return mapping['nsx_port_id']
    except exc.NoResultFound:
        pass


def get_mismatch_logical_ports(context, nsxlib, plugin, get_filters=None):
    neutron_ports = plugin.get_ports(context, filters=get_filters)

    # get pre-defined profile ids
    dhcp_profile_id = get_dhcp_profile_id(nsxlib)
    dhcp_profile_key = (
        core_resources.SwitchingProfileTypes.SWITCH_SECURITY)
    spoofguard_profile_id = get_spoofguard_profile_id(nsxlib)
    spoofguard_profile_key = (
        core_resources.SwitchingProfileTypes.SPOOF_GUARD)
    qos_profile_key = core_resources.SwitchingProfileTypes.QOS

    problems = []
    for port in neutron_ports:
        neutron_id = port['id']
        # get the network nsx id from the mapping table
        nsx_id = get_port_nsx_id(context.session, neutron_id)
        if not nsx_id:
            # skip external ports
            pass
        else:
            try:
                nsx_port = nsxlib.logical_port.get(nsx_id)
            except nsxlib_exc.ResourceNotFound:
                problems.append({'neutron_id': neutron_id,
                                 'nsx_id': nsx_id,
                                 'error': 'Missing from backend',
                                 'error_type': PORT_ERROR_TYPE_MISSING})
                continue

            # Port found on backend!
            # Check that it has all the expected switch profiles.
            # create a dictionary of the current profiles:
            profiles_dict = {}
            for prf in nsx_port['switching_profile_ids']:
                profiles_dict[prf['key']] = prf['value']

            # DHCP port: neutron dhcp profile should be attached
            # to logical ports created for neutron DHCP but not
            # for native DHCP.
            if (port.get('device_owner') == const.DEVICE_OWNER_DHCP and
                not cfg.CONF.nsx_v3.native_dhcp_metadata):
                prf_id = profiles_dict[dhcp_profile_key]
                if prf_id != dhcp_profile_id:
                    add_profile_mismatch(problems, neutron_id, nsx_id,
                                         prf_id, "DHCP security")

            # Port with QoS policy: a matching profile should be attached
            qos_policy_id = qos_utils.get_port_policy_id(context,
                                                         neutron_id)
            if qos_policy_id:
                qos_profile_id = nsx_db.get_switch_profile_by_qos_policy(
                    context.session, qos_policy_id)
                prf_id = profiles_dict[qos_profile_key]
                if prf_id != qos_profile_id:
                    add_profile_mismatch(problems, neutron_id, nsx_id,
                                         prf_id, "QoS")

            # Port with security & fixed ips/address pairs:
            # neutron spoofguard profile should be attached
            port_sec, has_ip = plugin._determine_port_security_and_has_ip(
                context, port)
            addr_pair = port.get(addr_apidef.ADDRESS_PAIRS)
            if port_sec and (has_ip or addr_pair):
                prf_id = profiles_dict[spoofguard_profile_key]
                if prf_id != spoofguard_profile_id:
                    add_profile_mismatch(problems, neutron_id, nsx_id,
                                         prf_id, "Spoof Guard")

            # Check the address bindings
            if port_sec:
                nsx_address_bindings = nsx_port.get('address_bindings', [])
                nsx_ips = [x['ip_address'] for x in nsx_address_bindings]
                nsx_macs = [x['mac_address'] for x in nsx_address_bindings]
                neutron_ips = [x['ip_address']
                               for x in port.get('fixed_ips', [])]
                neutron_mac = port['mac_address']
                different_macs = [mac for mac in nsx_macs
                                  if mac != neutron_mac]
                if (len(nsx_ips) != len(neutron_ips) or
                    set(nsx_ips) != set(neutron_ips)):
                    problems.append({'neutron_id': neutron_id,
                                     'nsx_id': nsx_id,
                                     'port': port,
                                     'error': 'Different IP address bindings',
                                     'error_type': PORT_ERROR_TYPE_BINDINGS})
                elif different_macs:
                    problems.append({'neutron_id': neutron_id,
                                     'nsx_id': nsx_id,
                                     'port': port,
                                     'error': 'Different MAC address bindings',
                                     'error_type': PORT_ERROR_TYPE_BINDINGS})
    return problems



# Copyright 2020 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.callbacks import registry

from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsx.shell.admin.plugins.nsxv3.resources import migration
from vmware_nsx.shell import resources as shell


@admin_utils.output_header
def cleanup_db_mappings(resource, event, trigger, **kwargs):
    """Delete all entries from nsx-t mapping tables in DB"""
    return migration.cleanup_db_mappings(resource, event, trigger, **kwargs)


registry.subscribe(cleanup_db_mappings,
                   constants.NSX_MIGRATE_T_P,
                   shell.Operations.CLEAN_ALL.value)



# Copyright 2019 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

LB_METHOD_ROUND_ROBIN = 'ROUND_ROBIN'
LB_METHOD_LEAST_CONNECTIONS = 'LEAST_CONNECTIONS'
LB_METHOD_SOURCE_IP = 'SOURCE_IP'
SUPPORTED_LB_ALGORITHMS = (LB_METHOD_LEAST_CONNECTIONS, LB_METHOD_ROUND_ROBIN,
                           LB_METHOD_SOURCE_IP)

PROTOCOL_TCP = 'TCP'
PROTOCOL_HTTP = 'HTTP'
PROTOCOL_HTTPS = 'HTTPS'
PROTOCOL_TERMINATED_HTTPS = 'TERMINATED_HTTPS'
POOL_SUPPORTED_PROTOCOLS = (PROTOCOL_TCP, PROTOCOL_HTTPS, PROTOCOL_HTTP)
LISTENER_SUPPORTED_PROTOCOLS = (PROTOCOL_TCP, PROTOCOL_HTTPS, PROTOCOL_HTTP,
                                PROTOCOL_TERMINATED_HTTPS)

LISTENER_POOL_COMPATIBLE_PROTOCOLS = (
    (PROTOCOL_TCP, PROTOCOL_TCP),
    (PROTOCOL_HTTP, PROTOCOL_HTTP),
    (PROTOCOL_HTTPS, PROTOCOL_HTTPS),
    (PROTOCOL_HTTP, PROTOCOL_TERMINATED_HTTPS))


HEALTH_MONITOR_PING = 'PING'
HEALTH_MONITOR_TCP = 'TCP'
HEALTH_MONITOR_HTTP = 'HTTP'
HEALTH_MONITOR_HTTPS = 'HTTPS'

SUPPORTED_HEALTH_MONITOR_TYPES = (HEALTH_MONITOR_HTTP, HEALTH_MONITOR_HTTPS,
                                  HEALTH_MONITOR_PING, HEALTH_MONITOR_TCP)

HTTP_METHOD_GET = 'GET'
HTTP_METHOD_HEAD = 'HEAD'
HTTP_METHOD_POST = 'POST'
HTTP_METHOD_PUT = 'PUT'
HTTP_METHOD_DELETE = 'DELETE'
HTTP_METHOD_TRACE = 'TRACE'
HTTP_METHOD_OPTIONS = 'OPTIONS'
HTTP_METHOD_CONNECT = 'CONNECT'
HTTP_METHOD_PATCH = 'PATCH'


SUPPORTED_HTTP_METHODS = (HTTP_METHOD_GET, HTTP_METHOD_HEAD, HTTP_METHOD_POST,
                          HTTP_METHOD_PUT, HTTP_METHOD_DELETE,
                          HTTP_METHOD_TRACE, HTTP_METHOD_OPTIONS,
                          HTTP_METHOD_CONNECT, HTTP_METHOD_PATCH)

# URL path regex according to RFC 3986
# Format: path = "/" *( "/" segment )
#         segment       = *pchar
#         pchar         = unreserved / pct-encoded / sub-delims / ":" / "@"
#         unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
#         pct-encoded   = "%" HEXDIG HEXDIG
#         sub-delims    = "!" / "$" / "&" / "'" / "(" / ")"
#                         / "*" / "+" / "," / ";" / "="
SUPPORTED_URL_PATH = (
    "^(/([a-zA-Z0-9-._~!$&\'()*+,;=:@]|(%[a-fA-F0-9]{2}))*)+$")

SESSION_PERSISTENCE_SOURCE_IP = 'SOURCE_IP'
SESSION_PERSISTENCE_HTTP_COOKIE = 'HTTP_COOKIE'
SESSION_PERSISTENCE_APP_COOKIE = 'APP_COOKIE'
SUPPORTED_SP_TYPES = (SESSION_PERSISTENCE_SOURCE_IP,
                      SESSION_PERSISTENCE_HTTP_COOKIE,
                      SESSION_PERSISTENCE_APP_COOKIE)

L7_RULE_TYPE_HOST_NAME = 'HOST_NAME'
L7_RULE_TYPE_PATH = 'PATH'
L7_RULE_TYPE_FILE_TYPE = 'FILE_TYPE'
L7_RULE_TYPE_HEADER = 'HEADER'
L7_RULE_TYPE_COOKIE = 'COOKIE'
SUPPORTED_L7_RULE_TYPES = (L7_RULE_TYPE_HOST_NAME,
                           L7_RULE_TYPE_PATH,
                           L7_RULE_TYPE_FILE_TYPE,
                           L7_RULE_TYPE_HEADER,
                           L7_RULE_TYPE_COOKIE)

L7_RULE_COMPARE_TYPE_REGEX = 'REGEX'
L7_RULE_COMPARE_TYPE_STARTS_WITH = 'STARTS_WITH'
L7_RULE_COMPARE_TYPE_ENDS_WITH = 'ENDS_WITH'
L7_RULE_COMPARE_TYPE_CONTAINS = 'CONTAINS'
L7_RULE_COMPARE_TYPE_EQUAL_TO = 'EQUAL_TO'
SUPPORTED_L7_RULE_COMPARE_TYPES = (L7_RULE_COMPARE_TYPE_REGEX,
                                   L7_RULE_COMPARE_TYPE_STARTS_WITH,
                                   L7_RULE_COMPARE_TYPE_ENDS_WITH,
                                   L7_RULE_COMPARE_TYPE_CONTAINS,
                                   L7_RULE_COMPARE_TYPE_EQUAL_TO)

L7_POLICY_ACTION_REJECT = 'REJECT'
L7_POLICY_ACTION_REDIRECT_TO_POOL = 'REDIRECT_TO_POOL'
L7_POLICY_ACTION_REDIRECT_TO_URL = 'REDIRECT_TO_URL'
SUPPORTED_L7_POLICY_ACTIONS = (L7_POLICY_ACTION_REJECT,
                               L7_POLICY_ACTION_REDIRECT_TO_POOL,
                               L7_POLICY_ACTION_REDIRECT_TO_URL)

URL_REGEX = "http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*(),]|\
             (?:%[0-9a-fA-F][0-9a-fA-F]))+"

# See RFCs 2616, 2965, 6265, 7230: Should match characters valid in a
# http header or cookie name.
HTTP_HEADER_COOKIE_NAME_REGEX = r'\A[a-zA-Z0-9!#$%&\'*+-.^_`|~]+\Z'

# See RFCs 2616, 2965, 6265: Should match characters valid in a cookie value.
HTTP_COOKIE_VALUE_REGEX = r'\A[a-zA-Z0-9!#$%&\'()*+-./:<=>?@[\]^_`{|}~]+\Z'

# See RFC 7230: Should match characters valid in a header value.
HTTP_HEADER_VALUE_REGEX = (r'\A[a-zA-Z0-9'
                           r'!"#$%&\'()*+,-./:;<=>?@[\]^_`{|}~\\]+\Z')

# Also in RFC 7230: Should match characters valid in a header value
# when quoted with double quotes.
HTTP_QUOTED_HEADER_VALUE_REGEX = (r'\A"[a-zA-Z0-9 \t'
                                  r'!"#$%&\'()*+,-./:;<=>?@[\]^_`{|}~\\]*"\Z')

STATS_ACTIVE_CONNECTIONS = 'active_connections'
STATS_MAX_CONNECTIONS = 'max_connections'
STATS_TOTAL_CONNECTIONS = 'total_connections'
STATS_CURRENT_SESSIONS = 'current_sessions'
STATS_MAX_SESSIONS = 'max_sessions'
STATS_TOTAL_SESSIONS = 'total_sessions'
STATS_IN_BYTES = 'bytes_in'
STATS_OUT_BYTES = 'bytes_out'
STATS_CONNECTION_ERRORS = 'connection_errors'
STATS_RESPONSE_ERRORS = 'response_errors'
STATS_STATUS = 'status'
STATS_HEALTH = 'health'
STATS_FAILED_CHECKS = 'failed_checks'

# Constants to extend status strings in neutron.plugins.common.constants
ONLINE = 'ONLINE'
OFFLINE = 'OFFLINE'
DEGRADED = 'DEGRADED'
DISABLED = 'DISABLED'
NO_MONITOR = 'NO_MONITOR'
OPERATING_STATUSES = (ONLINE, OFFLINE, DEGRADED, DISABLED, NO_MONITOR)

NO_CHECK = 'no check'

# LBaaS V2 Agent Constants
LBAAS_AGENT_SCHEDULER_V2_EXT_ALIAS = 'lbaas_agent_schedulerv2'
AGENT_TYPE_LOADBALANCERV2 = 'Loadbalancerv2 agent'
LOADBALANCER_PLUGINV2 = 'n-lbaasv2-plugin'
LOADBALANCER_AGENTV2 = 'n-lbaasv2_agent'

LOADBALANCER = "LOADBALANCER"
LOADBALANCERV2 = "LOADBALANCERV2"

# Used to check number of connections per second allowed
# for the LBaaS V1 vip and LBaaS V2 listeners. -1 indicates
# no limit, the value cannot be less than -1.
MIN_CONNECT_VALUE = -1

# LBaas V2 Table entities
LISTENER_EVENT = 'listener'
LISTENER_STATS_EVENT = 'listener_stats'
LOADBALANCER_EVENT = 'loadbalancer'
LOADBALANCER_STATS_EVENT = 'loadbalancer_stats'
MEMBER_EVENT = 'member'
OPERATING_STATUS = 'operating_status'
POOL_EVENT = 'pool'
